# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import json
import logging
import re
import subprocess
import sys
from collections import OrderedDict
from pathlib import Path
from textwrap import dedent

import git
import pytest

from ..cli import utils
from ..compat import metadata
from ..errors import ConfigurationError

_git_time = f"{42 * 365 * 24 * 3600} +0000"
_author = git.Actor("Bob Tester", "bob@example.net")
_commitargs = dict(
    author_date=_git_time,
    commit_date=_git_time,
    author=_author,
    committer=_author,
)


@pytest.mark.parametrize(
    "expected_args",
    (
        (
            "--extra-index-url",
            "https://test.pypi.org/simple/",
            "hopic>=1.19<2",
        ),
        (
            "--index-url",
            "https://test.pypi.org/simple/",
            "commisery>=0.2,<1",
        ),
        ("flake8",),  # noqa: E201
    ),
)
def test_install_extensions_from_multiple_indices(monkeypatch, run_hopic, expected_args):
    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        # del ['-c', constraints_file]
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", *expected_args]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    if expected_args[0] == "--extra-index-url":
        config = dedent(
            f"""\
                pip:
                  - with-extra-index:
                      - {expected_args[1]}
                    packages:
                      - {expected_args[-1]}
                """
        )
    elif expected_args[0] == "--index-url":
        config = dedent(
            f"""\
                pip:
                  - from-index: {expected_args[1]}
                    packages:
                      - {expected_args[-1]}
                """
        )
    else:
        assert not expected_args[0].startswith("-")
        config = dedent(
            f"""\
                pip: {json.dumps(expected_args)}
                """
        )
    (result,) = run_hopic(
        ("install-extensions",),
        config=config,
    )

    assert result.exit_code == 0


def test_with_single_extra_index(monkeypatch, run_hopic):
    extra_index = "https://test.pypi.org/simple/"
    pkg = "hopic>=1.19<2"

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        # del ['-c', constraints_file]
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, pkg]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    (result,) = run_hopic(
        ("install-extensions",),
        config=dedent(
            f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}
            """
        ),
    )

    assert result.exit_code == 0


def test_recursive_extension_installation(monkeypatch, run_hopic):
    extra_index = "https://test.pypi.org/simple/"
    pkg = "pipeline-template"
    template_pkg = "template-in-template"
    expected_pkg_install_order = [pkg, template_pkg]
    inner_template_called = []

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, expected_pkg_install_order.pop(0)]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    def template_template(volume_vars):
        inner_template_called.append(True)
        return dedent(
            """\
                        phases:
                          test:
                            variant:
                              - echo 'bob'
                        """
        )

    class TestTemplatePackage:
        name = template_pkg

        def load(self):
            return template_template

    def pipeline_template(volume_vars):
        return dedent(
            f"""\
                        pip:
                          - with-extra-index: {extra_index}
                            packages:
                              - {template_pkg}

                        config: !template {template_pkg}
                        """
        )

    class TestPipelinePackage:
        name = pkg

        def load(self):
            return pipeline_template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestPipelinePackage(), TestTemplatePackage())

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)

    (result,) = run_hopic(
        ("install-extensions",),
        config=dedent(
            f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}

                config: !template {pkg}
            """
        ),
    )

    assert result.exit_code == 0
    assert len(expected_pkg_install_order) == 0
    assert inner_template_called.pop()


def test_recursive_extension_installation_invalid_template_name(monkeypatch, run_hopic):
    extra_index = "https://test.pypi.org/simple/"
    pkg = "pipeline-template"
    template_pkg = "template-in-template"
    expected_pkg_install_order = [pkg, template_pkg]

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, expected_pkg_install_order.pop(0)]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    def pipeline_template(volume_vars):
        return dedent(
            f"""\
                        pip:
                          - with-extra-index: {extra_index}
                            packages:
                              - {template_pkg}

                        config: !template {template_pkg}-not-available
                        """
        )

    class TestPipelinePackage:
        def __init__(self):
            self.name = f"{pkg}"

        def load(self):
            return pipeline_template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestPipelinePackage(),)

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)

    (result,) = run_hopic(
        ("install-extensions",),
        config=dedent(
            f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}

                config: !template {pkg}
            """
        ),
    )

    assert result.exit_code == 0
    assert len(expected_pkg_install_order) == 0


@pytest.mark.parametrize("pass_through_environment_vars_template", (True, False))
def test_config_as_extension(monkeypatch, run_hopic, pass_through_environment_vars_template):
    """
    'config' item in hopic config yaml file should extend the existing config, not completely replace it.
    """
    extra_index = "https://test.pypi.org/simple/"
    template_pkg = "template-pkg-extension"
    test_image_name = "fake_docker_image"
    inner_template_called = []

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, template_pkg]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    def template(volume_vars):
        inner_template_called.append(True)
        template = ""
        if pass_through_environment_vars_template:
            template = dedent(
                """\
                  pass-through-environment-vars:
                    - THE_ENVIRONMENT_1
                """
            )

        template += dedent(
            """\
              phases:
                test:
                  variant:
                    - echo 'bob'
            """
        )

        return template

    class TestTemplatePackage:
        name = template_pkg

        def load(self):
            return template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestTemplatePackage(),)

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)

    hopic_ci_config = dedent(
        f"""\
            pip:
              - with-extra-index: {extra_index}
                packages:
                  - {template_pkg}

            image: {test_image_name}

            pass-through-environment-vars:
              - THE_ENVIRONMENT_1
              - THE_ENVIRONMENT_2

            config: !template {template_pkg}
        """
    )

    (install_ext_result,) = run_hopic(("install-extensions",), config=hopic_ci_config)
    assert install_ext_result.exit_code == 0
    assert inner_template_called.pop()

    (result,) = run_hopic(("show-config",), config=hopic_ci_config)
    output = json.loads(result.stdout, object_pairs_hook=OrderedDict)
    # validate that items from top-level configuration are not removed
    assert output["image"]["default"] == test_image_name
    # validate that items from config are added
    assert output["pass-through-environment-vars"] == ["THE_ENVIRONMENT_1", "THE_ENVIRONMENT_2"]
    assert "variant" in output["phases"]["test"]
    # validate that 'config' is actually removed to not trigger the process again
    assert "config" not in output


def test_config_has_duplicated_keys(monkeypatch, run_hopic):
    """
    'config' item only adds new items to the hopic config, if an item is already present, hopic should throw.
    """
    extra_index = "https://test.pypi.org/simple/"
    template_pkg = "template-pkg-duplicate-keys"
    test_image_name = "fake_docker_image"
    inner_template_called = []

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]

        assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, template_pkg]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    def template(volume_vars):
        inner_template_called.append(True)
        return dedent(
            """\
                        phases:
                          test:
                            variant:
                              - echo 'bob'
                        """
        )

    class TestTemplatePackage:
        def __init__(self):
            self.name = template_pkg

        def load(self):
            return template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestTemplatePackage(),)

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)

    (result,) = run_hopic(
        ("install-extensions",),
        config=dedent(
            f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {template_pkg}

                image: {test_image_name}

                phases:
                  test_custom:
                    variant_custom:
                      - echo 'jax'

                config: !template {template_pkg}
            """
        ),
    )
    assert isinstance(result.exception, ConfigurationError)
    assert inner_template_called.pop()
    err = result.exception.format_message()
    assert "top level configuration and 'config' item have duplicated keys: {'phases'}" in err


def test_invalid_template_name(run_hopic):
    """
    `hopic build` should return a clear error message when a template is specified
    that can't be found.
    """

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """
                config: !template xyzzy
            """
        ),
    )

    assert result.exit_code != 0
    assert any("No YAML template named 'xyzzy' available (props={})" in msg for _, msg in result.logs)


def test_recursive_extension_installation_version_functionality(monkeypatch, run_hopic):
    first_pkg = "firstorder"
    second_pkg = "secondorder"

    class FirstOrderTemplateFirst:
        def __init__(self):
            self.name = first_pkg

        @staticmethod
        def first_order_template_first_call(volume_vars):
            return dedent(
                f"""
                pip:
                  - packages:
                    - {second_pkg}

                phases:
                  yaml-error:
                    unknown:
                      - echo Unknown template
                """
            )

        def load(self):
            return self.first_order_template_first_call

    class FirstOrderTemplate:
        def __init__(self):
            self.name = first_pkg

        @staticmethod
        def first_order_template(volume_vars):
            return dedent(
                f"""
                pip:
                  - packages:
                    - {second_pkg}

                version:
                  tag: true
                  format: semver
                  bump:
                    policy: conventional-commits
                    strict: yes

                phases:
                  phase-one:
                    variant-one: !template {second_pkg}
                """
            )

        def load(self):
            return self.first_order_template

    class SecondOrderTemplate:
        name = second_pkg

        @staticmethod
        def second_order_template(volume_vars):
            return "- echo $PUBLISH_VERSION"

        def load(self):
            return self.second_order_template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        monkeypatch.setattr(metadata, "entry_points", lambda group: {"hopic.plugins.yaml": (FirstOrderTemplate(), SecondOrderTemplate())}[group])
        return (FirstOrderTemplateFirst(),)

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)
    monkeypatch.setattr(subprocess, "check_call", lambda *args, **kwargs: None)

    (result,) = run_hopic(
        ("build", "--dry-run"),
        config=dedent(
            f"""
            pip:
              - packages:
                - {first_pkg}

            config: !template {first_pkg}
            """
        ),
        tag="0.0.0",
    )

    assert result.exit_code == 0


def add_template(monkeypatch, pkg, config):
    def template(volume_vars):
        return config

    class TestPipelinePackage:
        name = pkg

        def load(self):
            return template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestPipelinePackage(),)

    monkeypatch.setattr(metadata, "entry_points", mock_entry_points)


def run_default_merge_flow(monkeypatch, run_hopic, config, pkg, message, target="master", merge_message=None):
    template_config = dedent(
        """\
                                version:
                                    format: semver
                                    tag: true
                                    bump:
                                        policy: conventional-commits
                                        strict: yes
                                    """
    )
    if merge_message is None:
        merge_message = message
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / "hopic-ci-config.yaml").open("w") as f:
            f.write(config)
        repo.index.add(("hopic-ci-config.yaml",))
        repo.index.commit(message="Initial commit", **_commitargs)
        repo.git.branch(target, move=True)
        repo.create_tag("0.0.0")

        # PR branch
        repo.head.reference = repo.create_head("something-useful")
        assert not repo.head.is_detached

        # Some change
        with (run_hopic.toprepo / "something.txt").open("w") as f:
            f.write("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message=message, **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", target),
        lambda: add_template(monkeypatch, pkg, template_config),
        (
            "prepare-source-tree",
            "--author-date",
            f"@{_git_time}",
            "--commit-date",
            f"@{_git_time}",
            "--author-name",
            _author.name,
            "--author-email",
            _author.email,
            "merge-change-request",
            "--source-remote",
            run_hopic.toprepo,
            "--source-ref",
            "something-useful",
            "--title",
            merge_message,
        ),
    )
    return result


@pytest.mark.parametrize(
    "merge_message, commit_message, expected_result_code",
    (
        ("feat: new feat", "fix: bump mismatch", 36),
        ("fix: a fix", "fix: bump mismatch", 0),
    ),
)
def test_extension_installation_version_config(monkeypatch, run_hopic, merge_message, commit_message, expected_result_code):
    extra_index = "https://test.pypi.org/simple/"
    pkg = "pipeline-template"
    expected_pkg_install_order = [pkg]

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]
        if len(expected_pkg_install_order):
            assert [*args] == [sys.executable, "-m", "pip", "install", "--extra-index-url", extra_index, expected_pkg_install_order.pop(0)]

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    result = run_default_merge_flow(
        monkeypatch,
        run_hopic,
        dedent(
            f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}

                config: !template {pkg}
            """
        ),
        pkg=pkg,
        message=commit_message,
        merge_message=merge_message,
    )

    assert result.exit_code == expected_result_code
    assert len(expected_pkg_install_order) == 0


def test_add_hopic_config_with_template_in_pr(capfd, monkeypatch, run_hopic):
    extra_index = "https://test.pypi.org/simple/"
    pkg = "pipeline-template"
    template_config = dedent(
        """\
                version:
                    format: semver
                    tag: true
                    bump:
                        policy: conventional-commits
                        strict: yes

                phases:
                    first-phase:
                        first-variant:
                            - echo 'Hello World!'
                """
    )
    subprocess_call = subprocess.check_call

    def mock_check_call(args, *popenargs, **kwargs):
        if "--user" in args:
            args.remove("--user")
        if "--verbose" in args:
            args.remove("--verbose")
        del args[4:6]
        if not all(arg in args for arg in ["pip", "install"]):
            subprocess_call(args)

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    merge_message = "ci: add hopic"
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        with (run_hopic.toprepo / "something.txt").open("w") as f:
            f.write("usable")
        repo.index.add(("something.txt",))
        repo.index.commit(message="ci: add hopic commit", **_commitargs)
        repo.create_tag("0.0.0")

        # PR branch
        repo.head.reference = repo.create_head("something-useful")
        assert not repo.head.is_detached

        # add hopic config
        with (run_hopic.toprepo / "hopic-ci-config.yaml").open("w") as f:
            f.write(
                dedent(
                    f"""\
                    pip:
                      - with-extra-index: {extra_index}
                        packages:
                          - {pkg}

                    config: !template {pkg}
                """
                )
            )
        repo.index.add(("hopic-ci-config.yaml",))
        repo.index.commit(message="ci: add hopic commit", **_commitargs)

    # Successful checkout and build
    (*_, result) = run_hopic(
        ("checkout-source-tree", "--target-remote", run_hopic.toprepo, "--target-ref", "master"),
        lambda: add_template(monkeypatch, pkg, template_config),
        (
            "prepare-source-tree",
            "--author-date",
            f"@{_git_time}",
            "--commit-date",
            f"@{_git_time}",
            "--author-name",
            _author.name,
            "--author-email",
            _author.email,
            "merge-change-request",
            "--source-remote",
            run_hopic.toprepo,
            "--source-ref",
            "something-useful",
            "--title",
            merge_message,
        ),
        ("build",),
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    out.splitlines()[-1] = "Hello World!"


@pytest.mark.parametrize("constraints", (True, False))
@pytest.mark.parametrize("upgrade", (True, False))
def test_extension_constraints_upgrade_mutual_exclusivity(monkeypatch, run_hopic, constraints, upgrade):
    def mock_check_call(args, *popenargs, **kwargs):
        if args[2:4] == ["pip", "install"]:
            if upgrade:
                assert "--upgrade" in args and "--constraints" not in args
            elif constraints:
                assert "--constraints" in args and "--upgrade" not in args
            else:
                assert "--constraints" not in args and "--upgrade" not in args

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    options = []
    if constraints:
        run_hopic.toprepo.mkdir()
        constraints_file = run_hopic.toprepo / "constraints.txt"
        constraints_file.write_text("test==1.2.3")
        options.extend(["--constraints", constraints_file.resolve()])
    if upgrade:
        options.append("--upgrade")

    (result,) = run_hopic(
        ("install-extensions", *options),
        config="",
    )
    if constraints and upgrade:
        assert isinstance(result.exception, click.BadOptionUsage)
        assert "mutually exclusive" in result.exception.format_message()
    else:
        assert result.exit_code == 0


def test_extension_constraints_functionality(monkeypatch, run_hopic, tmp_path):
    packages = (
        "commisery>=0.8.0",
        "somepackage==2.3.4",
        "someotherpackage==4.5.6",
    )
    constraint = "commisery==0.8.2"

    run_hopic.toprepo.mkdir()
    constraints_file = run_hopic.toprepo / "constraints.txt"
    constraints_file.write_text(constraint)

    options = ("--constraints", constraints_file.resolve())

    def mock_check_call(args, *popenargs, **kwargs):
        if args[2:4] == ["pip", "install"]:
            assert re.search(r"-c /[^ ]*/constraints.txt\b", " ".join(args)) is not None
            assert all(pkg in args for pkg in packages)

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    with monkeypatch.context() as write_text_m:
        # The first `Path.write_text` we encounter will be from `install-extensions`,
        # so check that our input constraint is logged.

        def mock_write_text(f, text, *args, **kwargs):
            assert constraint in text
            write_text_m.undo()

        write_text_m.setattr(Path, "write_text", mock_write_text)

        config = dedent(
            f"""\
            pip:
              - packages: {json.dumps(packages)}
            """
        )

        (_, result) = run_hopic(
            (
                "install-extensions",
                *options,
            ),
            ("build",),
            config=config,
        )

    assert result.exit_code == 0


@pytest.mark.parametrize(
    "installed_pip_version, expected_result",
    (
        ("21.2.0", 0),
        ("0.0.0", 1),
        (None, 2),
    ),
)
def test_extension_constraints_pip_version(monkeypatch, run_hopic, tmp_path, installed_pip_version, expected_result):
    TEST_PACKAGE = "pip"

    def mock_check(package):
        if package == TEST_PACKAGE:
            if not installed_pip_version:
                raise metadata.PackageNotFoundError
            return installed_pip_version
        return original_pkg_version(package)

    monkeypatch.setattr(subprocess, "check_call", lambda *x, **_: None)

    original_pkg_version = monkeypatch.setattr(utils, "get_package_version", mock_check)

    run_hopic.toprepo.mkdir()
    constraints_file = run_hopic.toprepo / "constraints.txt"
    constraints_file.touch()

    options = ("--constraints", constraints_file.resolve())
    config = f"pip: [ packages: [ {TEST_PACKAGE} ] ]"

    (result,) = run_hopic(
        ("install-extensions", *options),
        config=config,
    )

    if expected_result == 1:
        assert any(level == logging.WARNING and "upgrade your pip package" in line for level, line in result.logs)
    elif expected_result == 2:
        assert any(level == logging.WARNING and "is not installed" in line for level, line in result.logs)

    assert result.exit_code == 0



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
from textwrap import dedent
import typing

import pytest

from . import config_file
from .. import config_reader
from ..errors import ConfigurationError


@pytest.fixture
def mock_yaml_plugin(monkeypatch):
    class TestTemplate:
        name = 'example'

        def load(self):
            return self.example_template

        @staticmethod
        def example_template(
            volume_vars : typing.Mapping[str, str],
            *,
            required_param : str,
            optional_param: 'typing.Optional[str]' = None,
            defaulted_param : bool = False,
        ) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
            val = {
                'required': required_param,
                'defaulted': defaulted_param,
            }
            if optional_param is not None:
                val['optional'] = optional_param
            return (val,)

    class TestKwargTemplate:
        name = 'kwarg'

        def load(self):
            return self.kwarg_template

        @staticmethod
        def kwarg_template(
            volume_vars : typing.Mapping[str, str],
            *,
            required_param : str,
            **kwargs,
        ) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
            val = {
                'required': required_param,
                'defaulted': kwargs.get('defaulted-param', False),
            }
            optional = kwargs.pop('optional-param', None)
            if optional is not None:
                val['optional'] = optional
            val['kwargs'] = kwargs

            assert isinstance(val['required'], str)
            return (val,)

    class TestSimpleTemplate:
        name = 'simple'

        def load(self):
            return self.no_arg_template

        @staticmethod
        def no_arg_template(volume_vars):
            return ()

    class TestSequenceTemplate:
        name = 'sequence'

        def load(self):
            return self.sequence_template

        @staticmethod
        def sequence_template(
            volume_vars : typing.Mapping[str, str],
            *,
            sequence: typing.List[str] = [],
        ) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
            assert isinstance(sequence, typing.List)
            for v in sequence:
                assert isinstance(v, str)
            return ({'sequence': sequence},)

    class TestWrongDefaultTemplate:
        name = 'wrong-default'

        def load(self):
            return self.wrong_default_template

        @staticmethod
        def wrong_default_template(
            volume_vars : typing.Mapping[str, str],
            *,
            defaulted_param : 'str' = None,
        ) -> typing.Sequence[typing.Mapping[str, typing.Any]]:
            return ({
                'defaulted': defaulted_param,
            },)

    class TestWrongReturnTemplate:
        name = "wrong-return"

        def load(self):
            return self.wrong_return_template

        @staticmethod
        def wrong_return_template(
            volume_vars : typing.Mapping[str, str],
        ) -> typing.List[typing.Dict[str, typing.Any]]:
            return [
                {43: None},
            ]

    class TestNonGeneratorTemplate:
        name = "non-generator"

        def load(self):
            return self.non_generator_template

        @staticmethod
        def non_generator_template(
            volume_vars : typing.Mapping[str, str],
        ) -> typing.Generator:
            return [
                {"sh": ["ls"]},
            ]

    class TestGeneratorTemplate:
        name = 'generator'

        def load(self):
            return self.generator_template

        @staticmethod
        def generator_template(
            volume_vars : typing.Mapping[str, str],
            *,
            cmds: typing.List[str] = [],
        ) -> typing.Generator:
            yield "echo setup"
            yield from cmds
            yield "echo cleanup"

    class TestBadGeneratorTemplate:
        name = "bad-generator"

        def load(self):
            unwrapped = self.bad_generator_template
            while (
                hasattr(unwrapped, '__wrapped__')
                and getattr(unwrapped.__wrapped__, '__annotations__', None) is not None
                and getattr(unwrapped, '__annotations__') is unwrapped.__wrapped__.__annotations__
            ):
                unwrapped = unwrapped.__wrapped__
            return unwrapped

        @staticmethod
        def bad_generator_template(
            volume_vars : typing.Mapping[str, str],
        ) -> typing.Generator[typing.Mapping[str, typing.Any], None, None]:
            yield {"sh": ["ls"]}
            yield False

    def mock_entry_points():
        return {
            ep.name: ep
            for ep in [
                TestTemplate(),
                TestKwargTemplate(),
                TestSimpleTemplate(),
                TestSequenceTemplate(),
                TestWrongDefaultTemplate(),
                TestWrongReturnTemplate(),
                TestNonGeneratorTemplate(),
                TestGeneratorTemplate(),
                TestBadGeneratorTemplate(),
            ]
        }

    monkeypatch.setattr(config_reader, 'get_entry_points', mock_entry_points)


@pytest.mark.parametrize('version_build', (
    '',
    123,
    '@invalidmetadata',
    [1, 2, 3],
    (1,),
    {'name': 'value'},
    '+invalid',
    '.invalid',
))
def test_version_build_handles_invalid_values(version_build):
    with pytest.raises(ConfigurationError, match=r'version.build'):
        config_reader.read_version_info({}, {'build': version_build})


@pytest.mark.parametrize('version_build', (
    '1.70.0',
    '-oddbutvalid',
))
def test_version_build(version_build):
    result = config_reader.read_version_info({}, {'build': version_build})
    assert result['build'] == version_build


def test_version_build_non_semver():
    with pytest.raises(ConfigurationError, match=r'version.build'):
        config_reader.read_version_info({}, {'format': 'carver', 'build': '1.0.0'})


def test_default_version_bumping_config():
    version_info = config_reader.read_version_info(os.devnull, {})
    assert version_info["bump"]["policy"] == "disabled"


@pytest.mark.parametrize(
    "version_fields",
    (
        {"tag": True},
        {"tag": "v{version}"},
        {"file": "version.txt"},
    ),
    ids=lambda v: (v if isinstance(v, str) else json.dumps(v)),
)
def test_default_version_bumping_config_with_storage(version_fields):
    version_info = config_reader.read_version_info(os.devnull, {**version_fields})

    assert version_info["bump"]["policy"] == "constant"


def test_default_version_bumping_backwards_compatible_policy():
    version_info = config_reader.read_version_info(os.devnull, {"bump": "patch"})

    assert version_info["bump"]["policy"] == "constant"
    assert version_info["bump"]["field"] == "patch"


def test_environment_without_cmd():
    with pytest.raises(ConfigurationError, match=r"set 'environment' member .* doesn't have 'sh'"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - environment: {}
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_environment_type_mismatch():
    with pytest.raises(ConfigurationError, match=r"`environment\['sheep'\]` is not a string"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - environment:
                              sheep: 1
                            sh:
                              - printenv
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_environment_from_prefix():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
                phases:
                  test:
                    example:
                      - SHEEP=1 EMPTY= ./command.sh
                '''
            )
        ),
        {'WORKSPACE': None},
    )
    (out,) = cfg['phases']['test']['example']
    assert out['sh'] == ['./command.sh']
    assert dict(out['environment']) == {'SHEEP': '1', 'EMPTY': ''}


def test_node_label_type_mismatch():
    with pytest.raises(ConfigurationError, match=r"`node-label` .*? string .*? bool"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - node-label: true
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_node_label_mismatch():
    with pytest.raises(ConfigurationError, match=r"`node-label` .*?\bdiffers from .*?\bprevious.*?\bdefined"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      build:
                        example:
                          - node-label: first
                      test:
                        example:
                          - node-label: second
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_node_label_mismatch_single_phase():
    with pytest.raises(ConfigurationError, match=r"`node-label` .*?\bdiffers from .*?\bprevious.*?\bdefined"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      build:
                        example:
                          - node-label: first
                          - node-label: second
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_node_label_default_override():
    with pytest.raises(ConfigurationError, match=r"`node-label` .*?\boverride default"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      build:
                        example: []
                      test:
                        example:
                          - node-label: second
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_post_submit_node_label_mismatch():
    with pytest.raises(ConfigurationError, match=r"`node-label` .*?\bdiffers from .*?\bprevious.*?\bdefined"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    post-submit:
                      build:
                        - node-label: first
                      test:
                        - node-label: second
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_node_label_match():
    config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
                phases:
                  build:
                    example:
                      - node-label: first
                  test:
                    example:
                      - node-label: first
                '''
            )
        ),
        {'WORKSPACE': None},
    )


def test_post_submit_node_label_match():
    config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
                post-submit:
                  build:
                    - node-label: first
                  test:
                    - node-label: first
                '''
            )
        ),
        {'WORKSPACE': None},
    )


def test_template_reserved_param(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to use reserved keyword `volume-vars` to instantiate template `.*?`'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: example
                  required-param: PIPE
                  volume-vars: {}
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_missing_param(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?` without required parameter'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template "example"
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_mismatched_param_simple_type(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?`: type of required-param must be str; got bool instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: example
                  required-param: yes
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_mismatched_param_optional_type(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?`: type of optional-param must be .*?; got bool instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: example
                  optional-param: yes
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_snake_param(mock_yaml_plugin):
    with pytest.raises(
        ConfigurationError,
        match=r'(?i)trying to instantiate template `.*?` with unexpected parameter `required_param`.*? mean `required-param`',
    ):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: example
                  required_param: PIPE
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_unknown_param(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?` with unexpected parameter `unknown-param`'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: example
                  unknown-param: 42
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_without_optional_param(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: example
              required-param: PIPE
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out, = cfg['phases']['test']['example']
    assert 'optional' not in out


def test_template_with_explicitly_null_param(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: example
              required-param: PIPE
              optional-param: null
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out, = cfg['phases']['test']['example']
    assert 'optional' not in out


def test_template_with_optional_param(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: example
              required-param: PIPE
              optional-param: Have you ever seen the rain?
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out, = cfg['phases']['test']['example']
    assert 'optional' in out
    assert out['optional'] == 'Have you ever seen the rain?'


def test_template_kwargs_required_snake_param(mock_yaml_plugin):
    with pytest.raises(
        ConfigurationError,
        match=r'(?i)trying to instantiate template `.*?` with unexpected parameter `required_param`.*? mean `required-param`',
    ):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: kwarg
                  required_param: PIPE
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_kwargs_snake_param_in_kwarg(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: kwarg
              required-param: PIPE
              something_extra-that_is-ridiculous: yes
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out, = cfg['phases']['test']['example']
    assert out['kwargs'] == {'something_extra-that_is-ridiculous': True}


def test_template_kwargs_missing_param(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?` without required parameter'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template "kwarg"
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_kwargs_type_mismatch(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?`: type of required-param must be str; got NoneType instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: kwarg
                  required-param: null
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_simple_unknown_param(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?` with unexpected parameter `unknown-param`'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: simple
                  unknown-param: 42
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_simple_without_param(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template "simple"
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out = cfg['phases']['test']['example']
    assert out == []


def test_template_sequence_without_param(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template "sequence"
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out = cfg['phases']['test']['example'][0]['sequence']
    assert out == []


def test_template_with_wrong_default(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)\bwrong default of parameter for template `.*?`: type of defaulted-param must be str; got .*? instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: wrong-default
                  defaulted-param: mooh
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_sequence_with_single_entry(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: "sequence"
              sequence:
                - mooh
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    out = cfg['phases']['test']['example'][0]['sequence']
    assert out == ['mooh']


def test_template_sequence_with_str_instead_of_list(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?`: type of sequence must be\b.*?\blist; got str instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      test:
                        example: !template
                          name: "sequence"
                          sequence: mooh
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_template_sequence_with_type_mismatched_entry(mock_yaml_plugin):
    with pytest.raises(ConfigurationError, match=r'(?i)trying to instantiate template `.*?`: type of sequence\[1\] must be str; got bool instead'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template
                  name: "sequence"
                  sequence:
                    - mooh
                    - false
                    - sheep
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_wrong_return(mock_yaml_plugin):
    with pytest.raises((ConfigurationError, TypeError), match=r"(?i)return value(?:\[.*?\])? must be .*?; got .*? instead"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template wrong-return
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_non_generator(mock_yaml_plugin):
    with pytest.raises((ConfigurationError, TypeError), match=r"(?i)return value must be \S*\bGenerator; got list instead"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template non-generator
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_template_generator(mock_yaml_plugin):
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        phases:
          test:
            example: !template
              name: generator
              cmds:
                - echo "do something"
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    cmds = [cmd["sh"] for cmd in cfg["phases"]["test"]["example"]]
    assert cmds == [
        ["echo", "setup"],
        ["echo", "do something"],
        ["echo", "cleanup"],
    ]


def test_bad_generator_template(mock_yaml_plugin):
    with pytest.raises((ConfigurationError, TypeError), match=r"(?i)value yielded from generator\b.*?\bmust be (?:dict|\S*\bMapping); got bool instead"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              test:
                example: !template bad-generator
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_nested_command_list_flattening():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                phases:
                  test:
                    example:
                      - run-on-change: always
                      -
                        - description: something happening here
                        -
                            - echo "Tada!"
                            - sh: echo "Tada!"
                            - sh: [echo, "Tada!"]
                """
            )
        ),
        {'WORKSPACE': None},
    )
    out = [dict(cmd) for cmd in cfg['phases']['test']['example']]
    assert out == [
        {"run-on-change": config_reader.RunOnChange.always},
        {"description": "something happening here"},
        {"environment": {}, "sh": ["echo", "Tada!"]},
        {"environment": {}, "sh": ["echo", "Tada!"]},
        {"environment": {}, "sh": ["echo", "Tada!"]},
    ]


def test_wait_on_full_previous_phase_dependency_type_mismatch():
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` doesn't contain a boolean"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      y:
                        b:
                          - wait-on-full-previous-phase: noo
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_wait_on_full_previous_phase_dependency_without_previous_phase():
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` defined but there is no previous phase"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      x:
                        a:
                          - wait-on-full-previous-phase: no
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_wait_on_full_previous_phase_dependency_without_previous_variant():
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` disabled but previous phase `x` doesn't contain variant `c`"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      x:
                        a:
                          - {dep_option}
                        b:
                          - echo monkey
                      y:
                        b:
                          - wait-on-full-previous-phase: no
                        c:
                          - wait-on-full-previous-phase: no
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_wait_on_full_previous_phase_dependency_multiple_definitions():
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` defined multiple times"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      x:
                        a:
                          - {dep_option}
                        b:
                          - echo monkey
                      y:
                        b:
                          - wait-on-full-previous-phase: no
                          - wait-on-full-previous-phase: no
                    """
                )
            ),
            {'WORKSPACE': None},
        )


@pytest.mark.parametrize(
    "dep_option",
    (
        "stash: {includes: test/**}",
        "worktrees: {doc/build/html: {commit-message: bla bla}}",
    ),
)
def test_wait_on_full_previous_phase_dependency_violation(dep_option):
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` disabled but previous phase `x` uses dependency-creating options"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    phases:
                      x:
                        a:
                          - {dep_option}
                        b:
                          - echo monkey
                      y:
                        b:
                          - wait-on-full-previous-phase: no
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_wait_on_full_previous_phase_dependency_run_on_change():
    with pytest.raises(ConfigurationError, match=r"(?i)`wait-on-full-previous-phase` disabled but `y`.`a`.`run-on-change` set to a value other than always"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      x:
                        a:
                          - run-on-change: always
                      y:
                        a:
                          - run-on-change: only
                            wait-on-full-previous-phase: no
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_wait_on_full_previous_phase_dependency_similar_run_on_change():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                phases:
                    x:
                      a:
                        - run-on-change: only
                    y:
                      a:
                        - run-on-change: only
                          wait-on-full-previous-phase: no
                """
            ),
        ),
        {"WORKSPACE": None},
    )
    (x_a,) = cfg["phases"]["x"]["a"]
    (y_a,) = cfg["phases"]["y"]["a"]
    assert "run-on-change" in x_a
    assert x_a["run-on-change"] == config_reader.RunOnChange.only
    assert "run-on-change" in y_a
    assert y_a["run-on-change"] == config_reader.RunOnChange.only
    assert "wait-on-full-previous-phase" in y_a
    assert y_a["wait-on-full-previous-phase"] is False


def test_wait_on_full_previous_phase_dependency_default_yes():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                phases:
                  x:
                    b:
                      - touch monkey
                  y:
                    b:
                      - cat monkey
                    c:
                      - touch pig
                """
            )
        ),
        {'WORKSPACE': None},
    )
    (x_b,) = cfg['phases']['x']['b']
    (y_b,) = cfg['phases']['y']['b']
    (y_c,) = cfg['phases']['y']['c']
    assert 'wait-on-full-previous-phase' not in x_b
    assert y_b['wait-on-full-previous-phase'] is True
    assert 'wait-on-full-previous-phase' not in y_c


def test_docker_run_extra_arguments_forbidden_option():
    with pytest.raises(
        ConfigurationError,
        match="`extra-docker-args` member of `v-one` contains one or more options that are not allowed:",
    ) as exc:
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      p-one:
                        v-one:
                          - image: buildpack-deps:18.04
                            extra-docker-args:
                              hostname: TESTBAK
                              user: root
                              workspace: /dev
                          - echo This build shall fail
                    """
                )
            ),
            {"WORKSPACE": None},
        )

    err = exc.value.format_message()
    for option in ("user", "workspace"):
        assert option in err.splitlines()[2], f'expected {option} in error message'


def test_docker_run_extra_arguments_wrong_type():
    with pytest.raises(
        ConfigurationError,
        match=r"`(?:p-one\.v-one\.)?extra-docker-args` (?:is not a valid mapping of extra docker arguments: )?"
        r'(?:argument|type of dict item) [`"]hostname[`"] for (?:`v-one`|extra-docker-args) (?:should|must) be (?:a\b)?\s*str[,;] '
        r"(?:not a float|got float instead)",
    ):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    image:
                      default: buildpack-deps:18.04

                    phases:
                      p-one:
                        v-one:
                          - extra-docker-args:
                              hostname: 3.14
                          - echo This build shall fail
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_ci_locks():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
        ci-locks:
            - branch: master
              repo-name: FICTIONAL/some-lock
            - branch: master
              repo-name: FICTIONAL/some-other-lock
              lock-on-change: never
                '''
            )
        ),
        {'WORKSPACE': None}
    )
    ci_locks = cfg['ci-locks']
    assert isinstance(ci_locks, list)
    assert ci_locks[0]['lock-on-change'] == 'always'
    assert ci_locks[1]['lock-on-change'] == 'never'


def test_ci_locks_wrong_lock_on_change_value():
    with pytest.raises(ConfigurationError, match='has an invalid attribute "lock-on-change", expected one of'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            ci-locks:
            - branch: master
              repo-name: FICTIONAL/some-other-lock
              lock-on-change: never123
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_ci_locks_wrong_branch_value():
    with pytest.raises(ConfigurationError, match='has an invalid attribute "branch", expected a str, but got a list'):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            ci-locks:
            - branch:
              - master
              repo-name: FICTIONAL/some-other-lock
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_multiple_options_on_archive():
    with pytest.raises(ConfigurationError, match=r"are not allowed in the same [Aa]rchive configuration, use only 'allow-missing"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              build:
                example:
                  - archive:
                     artifacts: doesnotexist.txt
                     allow-missing: true
                     allow-empty-archive: true
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_allow_empty_archive_empty_variant_removed():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
            phases:
              build:
                example:
                  - archive:
                     artifacts: doesnotexist.txt
                     allow-empty-archive: true
                '''
            )
        ),
        {'WORKSPACE': None}
    )

    out = cfg['phases']['build']['example'][0]['archive']
    assert 'allow-missing' in out
    assert type(out['allow-missing']) is bool
    assert 'allow-empty-archive' not in out


def test_archive_allow_missing_not_boolean():
    with pytest.raises(ConfigurationError, match=r"'build.example.archive.allow-missing' should be a boolean, not a str"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              build:
                example:
                  - archive:
                     artifacts: doesnotexist.txt
                     allow-missing: 'true'
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_allow_empty_junit():
    with pytest.raises(ConfigurationError, match=r"JUnit configuration did not contain mandatory field 'test-results'"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              build:
                example:
                  - junit:
                     test: doesnotexist.txt
                     allow-missing: true
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_allow_failures_junit():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
        phases:
          build:
            example:
              - junit:
                 test-results: doesnotexist.txt
                 allow-failures: true
                """
            ),
        ),
        {"WORKSPACE": None},
    )

    out = cfg["phases"]["build"]["example"][0]["junit"]
    assert "allow-failures" in out
    assert out["allow-failures"] is True


def test_generated_config_has_test_results():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
            phases:
              build:
                example:
                  - junit:
                     doesnotexistjunitresult.xml
                '''
            )
        ),
        {'WORKSPACE': None}
    )

    out = cfg['phases']['build']['example'][0]['junit']
    assert 'test-results' in out


def test_junit_allow_missing_not_boolean():
    with pytest.raises(ConfigurationError, match=r"'build.example.junit.allow-missing' should be a boolean, not a str"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
            phases:
              build:
                example:
                  - junit:
                     test-results: doesnotexist.txt
                     allow-missing: 'true'
                    '''
                )
            ),
            {'WORKSPACE': None}
        )


def test_junit_allow_failures_not_boolean():
    with pytest.raises(ConfigurationError, match=r"'build.example.junit.allow-failures' should be a boolean, not a str"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
            phases:
              build:
                example:
                  - junit:
                     test-results: doesnotexist.txt
                     allow-failures: 'true'
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


def test_archive_type_mismatch():
    with pytest.raises(ConfigurationError, match=r"member is not a mapping"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - archive: null
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_archive_missing_artifacts():
    with pytest.raises(ConfigurationError, match=r"lacks the mandatory 'artifacts' member"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - archive: {}
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_archive_artifacts_missing_pattern():
    with pytest.raises(ConfigurationError, match=r"lacks the mandatory 'pattern' member"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - archive:
                              artifacts:
                                - {}
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_archive_artifacts_pattern_type_mismatch():
    with pytest.raises(ConfigurationError, match=r"pattern' is not a string"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - archive:
                              artifacts:
                                - pattern: null
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


@pytest.mark.parametrize('pattern', (
    "**/a**",
    "**/a(*)(*)",
))
def test_archive_artifacts_pattern_invalid_double_star(pattern):
    with pytest.raises(
        ConfigurationError,
        match=fr"pattern' value of '{re.escape(pattern)}' is not a valid glob pattern: .*? '\*\*' can only be an entire path component",
    ):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    phases:
                      test:
                        example:
                          - archive:
                              artifacts:
                                - pattern: {json.dumps(pattern)}
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_junit_pattern_invalid_double_star():
    with pytest.raises(ConfigurationError, match=r"value of '.*?' is not a valid glob pattern: .*? '\*\*' can only be an entire path component"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    phases:
                      test:
                        example:
                          - junit:
                            - "**/a**"
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_ci_locks_reference_invalid_phase():
    with pytest.raises(ConfigurationError, match=r"referenced phase in ci-locks \(non-existing-phase\) doesn't exist"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    ci-locks:
                        - branch: branch
                          repo-name: repo
                          from-phase-onward: non-existing-phase

                    phases:
                      existing-phase:
                        example:
                          - echo 'test'
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_ci_locks_reference_wait_on_full_previous_phase_variant():
    with pytest.raises(ConfigurationError, match=r"referenced phase in ci-locks \(phase-2\) refers to variant \(wait-on-full-previous-phase-variant\) "
                                                 r"that has wait-on-full-previous-phase disabled"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    ci-locks:
                        - branch: branch
                          repo-name: repo
                          from-phase-onward: phase-2

                    phases:
                      phase-1:
                        example:
                          - echo 'test'

                        wait-on-full-previous-phase-variant:
                          - echo 'disable waiting on previous phase'

                      phase-2:
                        example:
                          - echo 'test'

                        wait-on-full-previous-phase-variant:
                          - wait-on-full-previous-phase: no

                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_ci_locks_duplicate_identifier():
    with pytest.raises(ConfigurationError, match=r"ci-lock with repo-name 'repo' and branch 'branch' already exists, "
                                                 r"this would lead to a deadlock"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    '''\
                    ci-locks:
                        - branch: branch
                          repo-name: repo
                          from-phase-onward: phase-1
                        - branch: branch
                          repo-name: repo

                    phases:
                      phase-1:
                        example:
                          - echo 'test'
                    '''
                )
            ),
            {'WORKSPACE': None},
        )


def test_ci_locks_on_phase_forward():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                '''\
                ci-locks:
                  - branch: branch
                    repo-name: repo
                    from-phase-onward: phase-1

                phases:
                  phase-1:
                    example:
                      - echo 'test'
                '''
            )
        ),
        {'WORKSPACE': None},
    )
    out = cfg['ci-locks'][0]
    assert 'from-phase-onward' in out


@pytest.mark.parametrize("timeout", (-1, 0, "yes", "mooh"))
def test_non_positive_timeout(timeout):
    with pytest.raises(ConfigurationError, match=r"`timeout` member of `a\.x` must be a positive real number"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    phases:
                      a:
                        x:
                          - timeout: {timeout}
                            sh: echo mooh
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


def test_global_timeout_multiple_definitions():
    with pytest.raises(ConfigurationError, match=r"`a\.x\[1\]` .*?\bdefine global `timeout` multiple"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      a:
                        x:
                          - timeout: 1
                          - timeout: 2
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


def test_global_timeout_after_sh():
    with pytest.raises(ConfigurationError, match=r"`a\.x\[1\]` .*?\bdefine global `timeout` after\b.*? 'sh' .*?\balready .*?\bgiven"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                      a:
                        x:
                          - echo mooh
                          - timeout: 2
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


@pytest.mark.parametrize("global_timeout", (3, 4))
@pytest.mark.parametrize("location", ("phases", "post-submit"))
def test_global_timeout_less_or_equal_than_sum_of_local_timeouts(global_timeout, location):
    with pytest.raises(
        ConfigurationError,
        match=fr"(?:`a\.x|post-submit\.a)\[0\]\.timeout` \({global_timeout} seconds\) .*?\bnot greater than summed .*?\btimeouts\b",
    ):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    {location}:
                      a:
                        {"x:" if location == "phases" else ""}
                          - timeout: {global_timeout}
                          - timeout: 2
                            sh: echo mooh
                          - timeout: 2
                            sh: echo beeh
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


def test_modality_sh_array():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                modality-source-preparation:
                  ALPHA:
                    - sh: [touch, new-file.txt]
                """
            ),
        ),
        {"WORKSPACE": None},
    )

    (cmd, *_) = cfg["modality-source-preparation"]["ALPHA"]
    assert cmd["sh"] == ["touch", "new-file.txt"]


@pytest.mark.parametrize(
    "block, field, value",
    (
        ("modality-source-preparation", "run-on-change", config_reader.RunOnChange.only),
        ("post-submit", "stash", {"includes": "stash/stash.txt"}),
    ),
    ids=lambda v: (v if isinstance(v, str) else json.dumps(v)),
)
def test_rejected_cmd_fields(block, field, value):
    with pytest.raises(ConfigurationError, match=fr"`{re.escape(block)}`?\.`?ALPHA(?:\[0\])?` contains? (?:forbidden|not permit).*?\bfields?\b"
                                                 fr".*?\b{re.escape(field)}\b"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    {block}:
                      ALPHA:
                        - {field}: {json.dumps(value)}
                          sh: touch new-file.txt
                    """
                )
            ),
            {'WORKSPACE': None},
        )


@pytest.mark.parametrize(
    "block, field, value",
    (
        ("post-submit", "junit", "result.xml"),
    ),
    ids=lambda v: (v if isinstance(v, str) else json.dumps(v)),
)
def test_unsupported_cmd_fields(block, field, value):
    with pytest.raises(ConfigurationError, match=fr"`{re.escape(block)}`?\.`?ALPHA(?:\[0\])?` contains? \b(?:unsupport|not support).*?\bfields?\b"
                                                 fr".*?\b{re.escape(field)}\b"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    f"""\
                    {block}:
                      ALPHA:
                        - {field}: {json.dumps(value)}
                          sh: touch new-file.txt
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_post_submit_type_error():
    with pytest.raises(ConfigurationError, match=r"`post-submit` doesn't contain a mapping but a list"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                            post-submit:
                                - echo 'hello Bob'
                    """
                )
            ),
            {'WORKSPACE': None},
        )


def test_post_submit():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                post-submit:
                    some-phase:
                        - echo 'hello Bob'
                """
            )
        ),
        {'WORKSPACE': None},
    )

    (out,) = cfg["post-submit"]["some-phase"]
    assert out["sh"] == ["echo", "hello Bob"]


def test_finally_variant():
    cfg = config_reader.read(
        config_file(
            "test-hopic-config.yaml",
            dedent(
                """\
                phases:
                    some-phase:
                        some-variant:
                            - sh: echo 'hello Bob'
                              finally:
                                - "echo 'sh finally'"
                            - finally:
                                - echo "finally"
                """
            ),
        ),
        {"WORKSPACE": None},
    )
    out = [elem for cmd in cfg["phases"]["some-phase"]["some-variant"] if "finally" in cmd for elem in cmd["finally"]]
    assert out[0]["sh"] == ["echo", "sh finally"]
    assert out[1]["sh"] == ["echo", "finally"]


def test_finally_variant_forbidden_keyword():
    with pytest.raises(ConfigurationError, match=r"`some-phase.some-variant-finally\[0\]` contains forbidden fields stash"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                        some-phase:
                            some-variant:
                                - sh: echo 'hello Bob'
                                  finally:
                                    - sh: "echo 'sh finally'"
                                      stash: "invalid"
                    """
                ),
            ),
            {"WORKSPACE": None},
        )


def test_sh_command_after_global_finally_error():
    with pytest.raises(ConfigurationError, match=r"`some-phase.some-variant\[1\]` attempting to define sh list after the global finally"):
        config_reader.read(
            config_file(
                "test-hopic-config.yaml",
                dedent(
                    """\
                    phases:
                        some-phase:
                            some-variant:
                                - finally:
                                    - sh: "echo 'sh finally'"
                                - sh: "echo hello world"
                    """
                ),
            ),
            {"WORKSPACE": None},
        )



# Copyright (c) 2019 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import source_date_epoch
from .markers import (
        docker,
    )
from .. import credentials
from .. import config_reader
from ..errors import (
    ConfigurationError,
    MissingFileError,
    UnknownPhaseError,
    StepTimeoutExpiredError,
    VersioningError,
)

from datetime import datetime
from textwrap import dedent
from typing import Pattern
import functools
import json
import logging
import os
import pytest
import re
import shlex
import signal
import stat
import subprocess
import sys
import time
import typing

from ..compat import metadata

from dateutil.parser import parse as parse_date
from dateutil.tz import tzutc


def test_missing_manifest(run_hopic):
    with pytest.raises(FileNotFoundError, match=r'(?:i).*\bivy manifest\b.*/dependency_manifest.xml\b'):
        (_,) = run_hopic(
            ("build",),
            config='''\
image: !image-from-ivy-manifest {}

phases:
  build:
    test:
      - cat /etc/lsb-release
''',
        )


def test_all_phases_and_variants(monkeypatch, run_hopic):
    expected = [
        ('build', 'a'),
        ('build', 'b'),
        ('test', 'b'),
        ('test', 'a'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
phases:
  build:
    a:
      - build a
    b:
      - build b
  test:
    b:
      - test b
    a:
      - test a
'''),
    )
    assert result.exit_code == 0


def test_filtered_phases(monkeypatch, run_hopic):
    expected = [
        ('build', 'a'),
        ('test', 'a'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build", "--phase=build", "--phase=test"),
        config=dedent('''\
phases:
  build:
    a:
      - build a
  test:
    a:
      - test a
  deploy:
    a:
      - deploy a
'''),
    )
    assert result.exit_code == 0


def test_filtered_non_existing_phase(monkeypatch, run_hopic):
    expected = []

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build", "--phase=build", "--phase=does-not-exist"),
        config=dedent('''\
phases:
  build:
    a:
      - build a
  test:
    a:
      - test a
  deploy:
    a:
      - deploy a
'''),
    )

    assert isinstance(result.exception, UnknownPhaseError)
    assert result.exception.format_message() == "build does not contain phase(s): does-not-exist"


def test_filtered_variants(monkeypatch, run_hopic):
    expected = [
        ('build', 'a'),
        ('build', 'c'),
        ('build', 'd'),
        ('test', 'a'),
        ('test', 'c'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build", "--variant=a", "--variant=c", "--variant=d"),
        config=dedent('''\
phases:
  build:
    a:
      - build a
    b:
      - build b
    c:
      - build c
    d:
      - build d
  test:
    b:
      - test b
    a:
      - test a
    c:
      - test c
'''),
    )

    assert (logging.WARNING, "phase 'test' does not contain variant 'd'") in result.logs
    assert result.exit_code == 0


def test_global_image(monkeypatch, run_hopic):
    def mock_check_call(args, *popenargs, **kwargs):
        assert args[0] == 'docker'
        assert tuple(args[-2:]) == ('buildpack-deps:18.04', './a.sh')

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config='''\
image: buildpack-deps:18.04

phases:
  build:
    a:
      - ./a.sh
''',
    )
    assert result.exit_code == 0


def test_default_image(monkeypatch, run_hopic):
    expected = [
        ('buildpack-deps:18.04', './a.sh'),
        ('buildpack-deps:buster', './b.sh'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert args[0] == 'docker'
        assert tuple(args[-2:]) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config='''\
image:
  default: buildpack-deps:18.04
  b: buildpack-deps:buster

phases:
  build:
    a:
      - ./a.sh
    b:
      - ./b.sh
''',
    )
    assert result.exit_code == 0
    assert not expected


def test_null_image(monkeypatch, run_hopic):
    expected = [
        {"docker": True, "cmd": ('buildpack-deps:18.04', './a.sh', '123')},
        {"docker": False, "cmd": ('./b.sh',)},
        {"docker": False, "cmd": ('./c.sh',)},
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        expected_call = expected.pop(0)
        if expected_call['docker']:
            assert args[0] == 'docker'
        else:
            assert args[0] != 'docker'

        cmd = tuple(args[-len(expected_call['cmd']):])
        assert cmd == expected_call['cmd']

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config='''\
image:
  default: buildpack-deps:18.04
  b: null

phases:
  build:
    a:
      - ./a.sh 123
    b:
      - ./b.sh
    c:
      - image: null
      - ./c.sh
''',
    )
    assert result.exit_code == 0
    assert not expected


@pytest.mark.parametrize('expect_forward_tty, has_stderr, has_stdin, has_stdout', (
    (True,   True , True , True),
    (False,  True , True , False),
    (False,  True , False, True),
    (False,  False, True , False),
    (False,  False, False, False),
))
def test_docker_run_arguments(run_hopic, expect_forward_tty, has_stderr, has_stdin, has_stdout):
    expected_image_command = [
        ('buildpack-deps:18.04', './a.sh'),
    ]
    uid = 42
    gid = 4242

    class MockDockerSockStat:
        st_gid = 2323
        st_mode = stat.S_IFSOCK | 0o0660

    def mock_check_call(args, *popenargs, **kwargs):
        expected_docker_args = [
            '--cap-add=SYS_PTRACE', '--rm', '--volume=/etc/passwd:/etc/passwd:ro',
            '--volume=/etc/group:/etc/group:ro', '--workdir=/code',
            f"--volume={os.getcwd()}:/code",
            f"--volume={os.getcwd()}/.ci:/cfg",
            f"--env=SOURCE_DATE_EPOCH={source_date_epoch}",
            '--env=HOME=/home/sandbox', '--env=_JAVA_OPTIONS=-Duser.home=/home/sandbox',
            f"--user={uid}:{gid}",
            '--net=host', f"--tmpfs=/home/sandbox:exec,uid={uid},gid={gid}",
            '--volume=/var/run/docker.sock:/var/run/docker.sock',
            f"--group-add={MockDockerSockStat.st_gid}",
            re.compile(r'^--cidfile=.*'),
        ]
        if expect_forward_tty:
            expected_docker_args += ['--tty']

        assert args[0] == 'docker'
        assert args[1] == 'run'
        image_command_length = len(tuple(expected_image_command[0]))
        assert tuple(args[-image_command_length:]) == expected_image_command.pop(0)
        docker_argument_list = args[2:-image_command_length]

        for docker_arg in expected_docker_args:
            if isinstance(docker_arg, Pattern):
                assert any(docker_arg.match(arg) for arg in docker_argument_list)
                docker_argument_list = [arg for arg in docker_argument_list if not docker_arg.match(arg)]
            else:
                assert docker_arg in docker_argument_list
                docker_argument_list.remove(docker_arg)
        assert docker_argument_list == []

    def set_monkey_patch_attrs(monkeypatch):
        monkeypatch.setattr(os, 'getuid', lambda: uid)
        monkeypatch.setattr(os, 'getgid', lambda: gid)
        old_os_stat = os.stat
        monkeypatch.setattr(os, 'stat', lambda path: MockDockerSockStat() if path == '/var/run/docker.sock' else old_os_stat(path))
        monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
        monkeypatch.setattr(sys.stderr, 'isatty', lambda: has_stderr)
        monkeypatch.setattr(sys.stdin, 'isatty', lambda: has_stdin)
        monkeypatch.setattr(sys.stdout, 'isatty', lambda: has_stdout)

    (result,) = run_hopic(
        ("build",),
        config='''\
image:
  default: buildpack-deps:18.04

phases:
  build:
    a:
      - docker-in-docker: yes
      - ./a.sh
''',
        monkeypatch_injector=set_monkey_patch_attrs,
        tag="0.0.0",
    )
    assert result.exit_code == 0
    assert not expected_image_command


@pytest.mark.parametrize(
    "extra_args, expected_args",
    (
        (
            {"device": "/dev/special-test-device"},
            ("--device=/dev/special-test-device",),
        ),
        (
            {
                "add-host": (
                    "my-test-host:10.13.37.254",
                    "my-other-test-host:10.13.37.253",
                ),
            },
            (
                "--add-host=my-test-host:10.13.37.254",
                "--add-host=my-other-test-host:10.13.37.253",
            ),
        ),
        (
            {
                "hostname": "TESTBAK",
                "init": True,
                "device": (
                    "/dev/null",
                    "/dev/special-test-device",
                ),
                "add-host": "my-test-host:10.13.37.254",
                "dns": "9.9.9.9",
            },
            (
                "--hostname=TESTBAK",
                "--init",
                "--device=/dev/null",
                "--device=/dev/special-test-device",
                "--add-host=my-test-host:10.13.37.254",
                "--dns=9.9.9.9",
            ),
        ),
    ),
    ids=("single-device", "multiple-hosts", "all-options"),
)
def test_docker_run_extra_arguments(monkeypatch, run_hopic, extra_args: typing.Mapping[str, typing.Any], expected_args: typing.Tuple[str, ...]):
    def mock_check_call(args, *popenargs, **kwargs):
        cmd, arg = args[-2:]
        assert cmd == "echo"

        monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
        for argument in expected_args:
            if arg == "Should contain extra args":
                assert argument in args
            else:
                assert arg == "Should not contain extra args"
                assert argument not in args

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            image:
              default: buildpack-deps:18.04

            phases:
              p-one:
                v-one:
                  - extra-docker-args: {json.dumps(extra_args)}
                  - echo 'Should contain extra args'
                  - echo 'Should contain extra args'
                v-two:
                  - echo 'Should not contain extra args'
              p-two:
                v-one:
                  - echo 'Should not contain extra args'
            """
        ),
    )

    if result.exception is not None:
        raise result.exception
    assert result.exit_code == 0


def test_docker_run_extra_arguments_whitespace_in_option(run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            '''\
            image:
              default: buildpack-deps:18.04

            phases:
              p-one:
                v-one:
                  - extra-docker-args:
                      hostname: something --user root
                  - echo This build shall fail
            '''
        ),
    )

    assert isinstance(result.exception, ConfigurationError)
    err = result.exception.format_message()
    assert 'argument `hostname` for `v-one` contains whitespace, which is not permitted.' in err


def test_override_default_volume(run_hopic):
    global_source = '/somewhere/over/the/rainbow'
    local_source = '/platform/nine/and/three/quarters'

    expected = [
            f"--volume={global_source}:/code",
            f"--volume={local_source}:/code",
        ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert expected.pop(0) in args

    def set_monkey_patch_attrs(monkeypatch):
        monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
        monkeypatch.setattr(os, 'makedirs', lambda _: None)

    (result,) = run_hopic(
        ('build',),
        config=dedent(
            f"""\
            image: buildpack-deps:18.04

            volumes:
              - source: {global_source}
                target: /code

            phases:
              test:
                regular:
                  - echo 'Hello World!'

                awesomeness:
                  - volumes:
                      - source: {local_source}
                        target: /code
                    sh: echo 'Hello World!'
            """
        ),
        monkeypatch_injector=set_monkey_patch_attrs,
    )
    assert result.exit_code == 0
    assert not expected


def test_image_override_per_phase(monkeypatch, run_hopic):
    expected = [
        ('buildpack-deps:18.04', './build-a.sh'),
        ('buildpack-deps:buster', './build-b.sh'),
        ('test-image-a:latest', './test-a.sh'),
        ('test-image-b:bleeding-edge', './test-b.sh'),
        ('buildpack-deps:18.04', './deploy-a.sh'),
        ('buildpack-deps:buster', './deploy-b.sh'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert args[0] == 'docker'
        assert tuple(args[-2:]) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
            image:
              default: buildpack-deps:18.04
              b: buildpack-deps:buster

            phases:
              build:
                a:
                  - ./build-a.sh
                b:
                  - ./build-b.sh

              test:
                a:
                  - image: test-image-a:latest
                  - ./test-a.sh
                b:
                  - image: test-image-b:bleeding-edge
                  - ./test-b.sh

              deploy:
                a:
                  - ./deploy-a.sh
                b:
                  - ./deploy-b.sh

            '''),
    )
    assert result.exit_code == 0
    assert not expected


def test_image_override_per_step(monkeypatch, run_hopic):
    """
    Verify that, when switching between images, and the absence of an image, the WORKSPACE is set properly.
    """
    expected = [
        ('buildpack-deps:18.04', './build-a.sh', '--workspace=/code'),
        ('./build-b.sh', '--workspace=${PWD}'),
        ('buildpack-deps:buster', './build-c.sh', '--workspace=/code'),
        ('./build-d.sh', '--workspace=${PWD}'),
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        expectation = [arg.replace('${PWD}', os.getcwd()) for arg in expected.pop(0)]
        if len(expected) % 2:
            assert args[0] == 'docker'
            assert args[-3:] == expectation
        else:
            assert args == expectation

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
            image: buildpack-deps:18.04

            phases:
              build:
                a:
                  - ./build-a.sh --workspace=${WORKSPACE}
                  - image: null
                    sh: ./build-b.sh --workspace=${WORKSPACE}
                  - image: buildpack-deps:buster
                    sh: ./build-c.sh --workspace=${WORKSPACE}
                  - image: null
                    sh: ./build-d.sh --workspace=${WORKSPACE}
'''),
    )
    assert result.exit_code == 0
    assert not expected


@pytest.mark.parametrize('signum', (
    signal.SIGINT,
    signal.SIGTERM,
))
def test_docker_terminated(monkeypatch, run_hopic, signum):
    expected = [
            ('docker', 'run'),
            ('docker', 'stop'),
        ]

    cid = 'the-magical-container-id'

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args[:2]) == expected.pop(0)

        if args[:2] == ['docker', 'run']:
            for arg in args:
                m = re.match(r'^--cidfile=(.*)', arg)
                if not m:
                    continue
                with open(m.group(1), 'w') as f:
                    f.write(cid)
            os.kill(os.getpid(), signum)
        else:
            assert tuple(args) == ('docker', 'stop', cid)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    def signal_handler(signum, frame):
        assert False, f"Failed to handle signal {signum}"
    old_handler = signal.signal(signum, signal_handler)
    if old_handler == signal.default_int_handler:
        signal.signal(signum, old_handler)
    try:
        (result,) = run_hopic(
            ("build",),
            config=dedent('''\
            image: buildpack-deps:18.04

            phases:
              build:
                a:
                  - ./build-a.sh
            '''),
        )
    finally:
        signal.signal(signum, old_handler)

    assert result.exit_code == 128 + signum
    assert not expected


@docker
def test_container_with_env_var(run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
            image: buildpack-deps:18.04

            pass-through-environment-vars:
              - THE_ENVIRONMENT

            phases:
              build:
                test:
                  - docker-in-docker: yes
                  - printenv THE_ENVIRONMENT
            '''),
        env={'THE_ENVIRONMENT': 'The Real Environment!'},
    )
    assert result.exit_code == 0


@docker
def test_container_without_env_var(run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
            image: buildpack-deps:18.04

            pass-through-environment-vars:
              - THE_ENVIRONMENT

            phases:
              build:
                test:
                  - printenv THE_ENVIRONMENT
            '''),
        env={'THE_ENVIRONMENT': None},
    )
    assert result.exit_code != 0


def test_command_with_source_date_epoch(capfd, run_hopic):
    (result,) = run_hopic(
        ("build",),
        config='''\
phases:
  build:
    test:
      - printenv SOURCE_DATE_EPOCH
''',
    )
    assert result.exit_code == 0
    out, err = capfd.readouterr()
    assert out.strip() == str(source_date_epoch)


def test_command_with_deleted_env_var(run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            '''\
            phases:
              build:
                test:
                  - environment:
                      SOURCE_DATE_EPOCH: null
                    sh: printenv SOURCE_DATE_EPOCH
            '''
        ),
    )
    assert result.exit_code != 0


def test_command_with_branch_and_commit(capfd, run_hopic):
    (_, result) = run_hopic(
        ("checkout-source-tree", "--clean", "--target-remote", ".", "--target-ref", "master"),
        ("build",),
        config=dedent('''\
            phases:
              build:
                test:
                  - echo ${GIT_BRANCH}=${GIT_COMMIT}
            '''),
    )
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    checkout_commit = out.splitlines()[0]
    claimed_branch, claimed_commit = out.splitlines()[1].split('=')
    assert claimed_branch == 'master'
    assert claimed_commit == checkout_commit


def test_command_without_branch(capfd, run_hopic):
    fallback_branch = "unknown-local-branch"

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            phases:
              build:
                test:
                  - echo -n ${{GIT_BRANCH:-{fallback_branch}}}
            """
        ),
    )
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert out == fallback_branch


def test_fallback_variable(capfd, run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """\
            phases:
              build:
                test:
                  - echo -n ${NON_EXISTANT:-fallback}
            """
        ),
    )
    assert result.exit_code == 0

    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert out == "fallback"


def test_empty_variant(run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
        phases:
          build:
            test: []'''),
    )
    assert result.exit_code == 0


def test_embed_variants(monkeypatch, run_hopic):
    expected = [
        ('./a.sh', 'test_argument'),
        ('./b.sh',),
        ('./test-a.sh',),
        ('./test-b.sh',),
    ]
    generate_script_path = "generate-variants.py"

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected.pop(0)

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                phases:
                  build:
                    a:
                      - ./a.sh test_argument
                    b:
                      - ./b.sh

                  test: !embed
                    cmd: {generate_script_path}'''),
        files={generate_script_path: (
            dedent('''\
                #!/usr/bin/env python3
                print(\'\'\'a:
                  - ./test-a.sh
                b:
                  - ./test-b.sh \'\'\')'''),
            lambda fname: os.chmod(fname, os.stat(fname).st_mode | stat.S_IEXEC))})
    assert result.exit_code == 0
    assert not expected


def test_embed_variants_syntax_error(capfd, run_hopic):
    generate_script_path = "generate-variants.py"

    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                phases:
                  test: !embed
                    cmd: {generate_script_path}
                '''),
        files={generate_script_path: (dedent('''\
                #!/usr/bin/env python3
                print(\'\'\'a:
                  - ./a-test.sh
                b:
                  - ./b-test.sh
                yaml_error \'\'\')
                '''), lambda fname: os.chmod(fname, os.stat(fname).st_mode | stat.S_IEXEC))})
    assert result.exit_code == 42
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert 'An error occurred when parsing the hopic configuration file' in out


# fmt: off
@pytest.mark.parametrize(
    "format, init_version, build, commit_count, dirty, expected_version, expected_pure_version, expected_debversion",
    (
        ("semver", "0.0.0"       , None   , 0, False, "0.0.0+g{commit}"                    , "0.0.0"                          , "0.0.0+g{commit}"                   ),
        ("semver", "1.2.3"       , None   , 2, False, "1.2.4-2+g{commit}"                  , "1.2.4-2"                        , "1.2.4~2+g{commit}"                 ),
        ("semver", "2.0.0"       , None   , 0, True , "2.0.1-0.dirty.{timestamp}+g{commit}", "2.0.1-0.dirty.{timestamp}"      , "2.0.1~0+dirty{timestamp}+g{commit}"),
        ("semver", "2.5.1"       , None   , 1, True , "2.5.2-1.dirty.{timestamp}+g{commit}", "2.5.2-1.dirty.{timestamp}"      , "2.5.2~1+dirty{timestamp}+g{commit}"),
        ("semver", "0.0.0"       , "1.0.0", 0, False, "0.0.0+g{commit}"                    , "0.0.0"                          , "0.0.0+g{commit}"                   ),
        ("carver", "1.2.3+PI42.3", None   , 0, False, "1.2.3+PI42.3"                       , "1.2.3+PI42.3"                   , "1.2.3+PI42.3"                      ),
        ("carver", "1.2.3+PI42.3", None   , 2, False, "1.2.3-2+PI42.4"                     , "1.2.3-2+PI42.4"                 , "1.2.3~2+PI42.4"                    ),
    )
)
# fmt: on
def test_version_variables_content(
    capfd,
    run_hopic,
    format,
    init_version,
    build,
    commit_count,
    dirty,
    expected_version,
    expected_pure_version,
    expected_debversion,
):
    (result,) = run_hopic(
        ("build",),
        config=dedent(f"""\
                version:
                  format: {format}
                  tag:    true
                  bump:   {'fix' if format == 'carver' else 'patch'}
                {('  build: ' + build) if build else ''}

                phases:
                  test:
                    version:
                      - echo ${{VERSION}}
                      - sh -c 'echo $${{VERSION}}'
                      - echo ${{PURE_VERSION}}
                      - sh -c 'echo $${{PURE_VERSION}}'
                      - echo ${{DEBVERSION}}
                      - sh -c 'echo $${{DEBVERSION}}'
                """),
        tag=init_version,
        commit_count=commit_count,
        dirty=dirty,
    )
    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    # Length needs to match the length of a commit hash of `git describe`
    commit_hash = str(result.commit)[:14]

    expected_version = expected_version.format(commit=commit_hash, timestamp='19700108000000')
    expected_pure_version = expected_pure_version.format(commit=commit_hash, timestamp='19700108000000')
    expected_debversion = expected_debversion.format(commit=commit_hash, timestamp='19700108000000')

    assert out.splitlines()[0] == expected_version
    assert out.splitlines()[0] == out.splitlines()[1]
    assert out.splitlines()[2] == expected_pure_version
    assert out.splitlines()[2] == out.splitlines()[3]
    assert out.splitlines()[4] == expected_debversion
    assert out.splitlines()[4] == out.splitlines()[5]


def test_execute_list(monkeypatch, run_hopic):
    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == ('echo', "an argument, with spaces and ' quotes", 'and-another-without')

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
                phases:
                  build:
                    a:
                      - sh: ['echo', "an argument, with spaces and ' quotes", 'and-another-without']
                '''),
    )
    assert result.exit_code == 0


def test_with_credentials_keyring_variable_names(monkeypatch, run_hopic, capfd):
    username = 'test_username'
    password = 'super_secret'
    credential_id = 'test_credentialId'
    project_name = 'test_project'

    def get_credential_id(project_name_arg, cred_id):
        assert credential_id == cred_id
        assert project_name == project_name_arg
        return username, password

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                project-name: {project_name}
                phases:
                  build_and_test:
                    clang-tidy:
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                      - sh -c "echo $USERNAME $PASSWORD"
                      - sh -c "echo $$USERNAME $$PASSWORD"
                    coverage:
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                          username-variable: 'TEST_USER'
                          password-variable: 'TEST_PASSWORD'
                      - sh -c "echo $TEST_USER $TEST_PASSWORD"
                      - sh -c "echo $$TEST_USER $$TEST_PASSWORD"
                '''),
    )
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out.splitlines()[0] == f'{username} {password}'
    assert out.splitlines()[1] == f'{username} {password}'
    assert out.splitlines()[2] == f'{username} {password}'
    assert out.splitlines()[3] == f'{username} {password}'
    assert result.exit_code == 0


@pytest.mark.parametrize('username, expected_username, password, expected_password', (
    ('test_username', None,                 '$&+,/:;=?@', '%24%26%2B%2C%2F%3A%3B%3D%3F%40'),
    ('señor_tester', 'se%C3%B1or_tester'  , 'password'  , None),
    ('señor_tester', 'se%C3%B1or_tester'  , '$&+,/:;=?@', '%24%26%2B%2C%2F%3A%3B%3D%3F%40'),
    ('señor tester', 'se%C3%B1or%20tester', 'the secret', 'the%20secret'),
))
def test_with_credentials_with_url_encoding(monkeypatch, run_hopic, capfd, username, expected_username, password, expected_password):
    if expected_username is None:
        expected_username = username
    if expected_password is None:
        expected_password = password
    credential_id = 'test_credentialId'
    project_name = 'test_project'

    def get_credential_id(project_name_arg, cred_id):
        assert credential_id == cred_id
        assert project_name == project_name_arg
        return username, password

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                project-name: {project_name}
                phases:
                  build_and_test:
                    clang-tidy:
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                          encoding: url
                      - sh -c "echo $USERNAME $PASSWORD"
                '''),
    )
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out.splitlines()[0] == f'{expected_username} {expected_password}'
    assert result.exit_code == 0


def test_dry_run_build(monkeypatch, run_hopic):
    template_build_command = ['build b from template']

    expected = [
        ['https://test.pypi.org/simple/', 'test_template>=42.42'],
        ['[dry-run] would execute:'],
        ['generate doc/build/html/output.txt'],
        template_build_command,
        ['docker run', '/test/dir:/tmp', 'test-image:42.42 invalid command a'],
        ['invalid command b'],
    ]

    def mock_check_call(args, *popenargs, **kwargs):
        assert 'https://test.pypi.org/simple/' in args and 'test_template>=42.42' in args

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    def mock_load_template(*args, **kwargs):
        return [{
            'sh': template_build_command
        }]
    monkeypatch.setattr(config_reader, 'load_yaml_template', mock_load_template)

    (result,) = run_hopic(
        ("build", "--dry-run"),
        config=dedent('''\
pip:
  - with-extra-index:
      - 'https://test.pypi.org/simple/'
    packages:
      - test_template>=42.42

phases:
  build:
    a:
      - worktrees:
          doc/build/html:
            commit-message: "Update documentation"

      - generate doc/build/html/output.txt
    b: !template 'test_template'
  test:
    a:
      - image: test-image:42.42
        volumes:
          - ./test/dir:/tmp
      - invalid command a
    b:
      - invalid command b
'''),
    )
    assert result.exit_code == 0

    for level, msg in result.logs:
        if level < logging.INFO:
            continue
        for expected_string in expected.pop(0):
            assert expected_string in msg
    assert not expected


def test_dry_run_does_not_ask_for_credentials(monkeypatch, run_hopic):
    def get_credential_id(project_name_arg, cred_id):
        assert False, "`get_credential_id` should not have been called in a dry run"

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build", "-n"),
        config=dedent('''\
                project-name: dummy
                phases:
                  p1:
                    v1:
                      - with-credentials:
                        - id: my-credentials
                          type: username-password
                      - sh -c 'echo $USERNAME $PASSWORD'
                      - echo $USERNAME $PASSWORD
                '''),
    )

    assert result.exit_code == 0
    assert result.logs[-2][1] == "sh -c 'echo ${USERNAME} ${PASSWORD}'"
    assert result.logs[-1][1] == "echo '${USERNAME}' '${PASSWORD}'"


def test_config_recursive_template_build(monkeypatch, run_hopic):
    extra_index = 'https://test.pypi.org/simple/'
    pkg = 'pipeline-template'
    template_pkg = 'template-in-template'
    expected_check_calls = [['pip', 'install', pkg], ['pip', 'install', template_pkg], ['echo', 'bob']]

    def mock_check_call(args, *popenargs, **kwargs):
        assert all(elem in args for elem in expected_check_calls.pop(0))

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    def template_template(volume_vars):
        return dedent("""\
                      config:
                        phases:
                          test:
                            variant:
                              - echo 'bob'
                        """)

    class TestTemplatePackage:
        name = template_pkg

        def load(self):
            return template_template

    def pipeline_template(volume_vars):
        return dedent(f"""\
                        pip:
                          - with-extra-index: {extra_index}
                            packages:
                              - {template_pkg}

                        config: !template {template_pkg}
                        """)

    class TestPipelinePackage:
        name = pkg

        def load(self):
            return pipeline_template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestPipelinePackage(), TestTemplatePackage())

    monkeypatch.setattr(metadata, 'entry_points', mock_entry_points)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}

                config: !template {pkg}
                """),
    )

    assert result.exit_code == 0
    assert expected_check_calls == []


def test_build_list_yaml_template(monkeypatch, run_hopic):
    extra_index = 'https://test.pypi.org/simple/'
    pkg = 'variant-template'
    expected_check_calls = [['pip', 'install', pkg], ['echo', 'bob the builder'], ['echo', 'second command']]

    def mock_check_call(args, *popenargs, **kwargs):
        assert all(elem in args for elem in expected_check_calls.pop(0))

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)

    def pipeline_template(volume_vars):
        return dedent("""\
                        - echo 'bob the builder'
                        - echo 'second command'
                        """)

    class TestPipelinePackage:
        def __init__(self):
            self.name = f'{pkg}'

        def load(self):
            return pipeline_template

    def mock_entry_points(*, group: str):
        assert group == "hopic.plugins.yaml"
        return (TestPipelinePackage(),)

    monkeypatch.setattr(metadata, 'entry_points', mock_entry_points)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f"""\
                pip:
                  - with-extra-index: {extra_index}
                    packages:
                      - {pkg}

                phases:
                  phase-1:
                    variant-1: !template {pkg}
                """),
    )

    assert result.exit_code == 0
    assert expected_check_calls == []


def test_with_credentials_obfuscation(monkeypatch, capfd, run_hopic):
    username = 'test_username'
    password = '\'#$%123'
    credential_id = 'test_credentialId'
    project_name = 'test_project'

    def get_credential_id(project_name_arg, cred_id):
        assert cred_id == credential_id
        assert project_name_arg == project_name
        return username, password

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                project-name: {project_name}
                phases:
                  build_and_test:
                    clang-tidy:
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                      - echo $USERNAME $PASSWORD
                '''),
    )
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert out.splitlines()[0] == f'{username} {password}'
    assert any("'${USERNAME}' '${PASSWORD}'" in msg for _, msg in result.logs)
    assert result.exit_code == 0


def test_with_missing_credentials_obfuscation(monkeypatch, capfd, run_hopic):
    credential_id = 'test_credentialId'
    project_name = 'test_project'

    def get_credential_id(*_):
        return None

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                project-name: {project_name}
                phases:
                  build_and_test:
                    clang-tidy:
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                      - echo Command without credential to verify missing credential(s)
                      - with-credentials:
                        - id: {credential_id}
                          type: username-password
                      - echo $USERNAME $PASSWORD
                '''),
    )
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    assert result.exit_code != 0


def test_with_credentials_obfuscation_empty_credentials(monkeypatch, capfd, run_hopic):
    def get_credential_id(project_name_arg, cred_id):
        return '', ''

    monkeypatch.setattr(credentials, 'get_credential_by_id', get_credential_id)

    (result,) = run_hopic(
        ("build",),
        config=dedent('''\
                project-name: dummy
                phases:
                  p1:
                    v1:
                      - with-credentials:
                        - id: some_empty_credentials
                          type: username-password
                      - echo $USERNAME $PASSWORD
                '''),
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out.splitlines()[0] == ' '


@pytest.mark.parametrize('version_config, expected_msg', (
    ('tag: true', r"^Failed to determine the current version from Git tag\."),
    ('file: version.txt', r"^Failed to determine the current version from file\."),
    ('bump: false', r"^Failed to determine the current version\."),
))
def test_version_variable_with_undetermined_version(capfd, run_hopic, version_config, expected_msg):
    (result,) = run_hopic(
        ("build",),
        config=dedent(f'''\
                version:
                  {version_config}
                phases:
                  phase-one:
                    variant-one:
                      - sh -c "set +u; echo VERSION=$$VERSION"
                      - echo $VERSION
                '''),
        tag=None,
    )

    assert isinstance(result.exception, VersioningError)
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out.splitlines()[0] == 'VERSION='
    assert re.search(expected_msg, result.exception.format_message(), re.MULTILINE)


@pytest.mark.parametrize("expected_hash, mtime, include_file", (
    ("c63b283df45487cb0d957e0aa799b9c72f78e45b707da6c3946701a63a514713", "2038-01-19 03:14:08.134210", "include/something/here.hpp"),
    ("a9ffb501db85f6bd1d5544ad2761c1d55449d8d36b9388c7798525a98eebdfe8", "1970-01-03 17:30:01.372010 +0000", "include/something/here.hpp"),
    ("cbbeedaa870244a7d7cdc655ad7ac890d7817f2b0172ca18203c8954b24050c4", "2038-01-19 03:14:08.134210", "include/from-really-long-directory-that-cannot-be-represented-all-that-super-well-in-the-super-teeny-tiny-little-amount-that-is-a-hundred-and-fifty-five-bytes/here.hpp"),  # noqa: E501
))
def test_normalize_artifacts(capfd, expected_hash, include_file, mtime, run_hopic):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            version:
              tag: true

            phases:
              a:
                x:
                  - archive:
                      artifacts: archive-${{PURE_VERSION}}.tar.gz
                  - mkdir -p {os.path.dirname(include_file)} src
                  - touch -d '{mtime}' {include_file} src/here.cpp
                  - tar czf archive-${{PURE_VERSION}}.tar.gz --format=pax include src
              b:
                x:
                  - sha256sum -b archive-${{PURE_VERSION}}.tar.gz
            """
        ),
        tag="0.0.0",
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    assert out == f"{expected_hash} *archive-0.0.0.tar.gz\n", "archive's hash should not depend on build time"


def test_normalize_dpkg(capfd, run_hopic):
    mtime = "2012-12-12 12:12:12.121212 +1200"
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            version:
              tag: true

            phases:
              a:
                x:
                  - archive:
                      artifacts: cow-mode_${{DEBVERSION}}_all.deb
                  - sh -c 'md5sum usr/share/doc/cows/special-cow-mode.txt > control/md5sums'
                  - tar -C control -czf ${{WORKSPACE}}/control.tar.gz md5sums control
                  - touch -d {shlex.quote(mtime)} usr/share/doc/cows/special-cow-mode.txt
                  - tar -czf data.tar.gz --format=pax usr
                  - ar rcU cow-mode_${{DEBVERSION}}_all.deb debian-binary control.tar.gz data.tar.gz
              b:
                x:
                  - ar x cow-mode_${{DEBVERSION}}_all.deb
                  - sha256sum -b debian-binary control.tar.gz data.tar.gz cow-mode_${{DEBVERSION}}_all.deb
            """
        ),
        files={
            "debian-binary": "2.0\n",
            "control/control": dedent(
                """
                Architecture: all
                Depends: apt, dpkg
                Recommends: aptitude
                Description: Adds Super Cow Powers to apt, even aptitude which claims not to have it.
                Maintainer: no-reply@example.com
                Package: cow-mode
                Priority: optional
                Version: 12.12.12
                Installed-Size: 27
                """
            ),
            "usr/share/doc/cows/special-cow-mode.txt": "Super Cow Powers unlocked!\n",
        },
        tag="12.12.12",
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)
    checksum = out.splitlines()[-4:]
    assert checksum == [
        "d526eb4e878a23ef26ae190031b4efd2d58ed66789ac049ea3dbaf74c9df7402 *debian-binary",
        "201071bdac6fd3fdeeedb4a37733deb4be3dbe94c0c4ae8aed5b2d01cde46f2c *control.tar.gz",
        "0fc7da07a77e902fe747544f2e65dc00998a015eeca4799636a985f52fad9fb7 *data.tar.gz",
        "92fccd82a96e65692edeaa2961e4d742c0087da46cb566d75f0fb9d2c5147a44 *cow-mode_12.12.12+gf7c740f27d9fdf_all.deb",
    ], "archive's hash should not depend on build time"


@pytest.mark.parametrize("archive_key", (
    "archive",
    "fingerprint",
    "junit",
))
def test_complain_about_missing_artifacts(run_hopic, archive_key):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            phases:
              a:
                x:
                  - {archive_key}:
                      {'test-results' if archive_key == 'junit' else 'artifacts'}: {archive_key}-doesnotexist.txt
            """
        ),
    )

    assert isinstance(result.exception, MissingFileError)
    msg = result.exception.format_message()
    assert re.search(r"\b[Nn]one of these mandatory .*? patterns matched a file\b", msg)
    assert f"{archive_key}-doesnotexist.txt" in msg


@pytest.mark.parametrize("archive_key", (
    "archive",
    "fingerprint",
    "junit",
))
def test_accept_present_artifacts(run_hopic, archive_key):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            phases:
              a:
                x:
                  - {archive_key}:
                      {'test-results' if archive_key == 'junit' else 'artifacts'}: {archive_key}-exists.txt
                    sh: touch {archive_key}-exists.txt
            """
        ),
    )

    assert result.exit_code == 0


@pytest.mark.parametrize("archive_key", (
    "archive",
    "fingerprint",
))
def test_permit_missing_artifacts(run_hopic, archive_key):
    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            phases:
              a:
                x:
                  - {archive_key}:
                      artifacts: {archive_key}-doesnotexist.txt
                      allow-missing: yes
                    junit:
                      test-results: junit-doesnotexist.xml
                      allow-missing: yes
            """
        ),
    )

    assert result.exit_code == 0


def test_build_times(capfd, run_hopic):
    expected_time = datetime.utcfromtimestamp(int(source_date_epoch)).replace(tzinfo=tzutc())
    expected_duration = 42 * 60 + 42.42

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """\
            phases:
              a:
                x:
                  - echo ${GIT_COMMIT_TIME} ${BUILD_DURATION}
            """
        ),
        env=dict(
            SOURCE_DATE_EPOCH=str(source_date_epoch + expected_duration),
        ),
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    git_commit_time, duration = out.split()
    git_commit_time = parse_date(git_commit_time)
    duration = float(duration)

    assert git_commit_time == expected_time
    assert duration == expected_duration


def test_build_identifiers(capfd, run_hopic):
    repo = "something"
    branch = "release/1"
    pr_number = 123
    job_build_number = 42

    expected_build_name = f"{repo}/{branch}"
    expected_build_number = f"PR-{pr_number} {job_build_number}"
    expected_build_url = f"https://some-where-jenkins.example.com/job/{repo}/job/PR-{pr_number}/{job_build_number}/"

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """\
            phases:
              a:
                x:
                  - echo ${BUILD_NAME}
                  - echo ${BUILD_NUMBER}
                  - echo ${BUILD_URL}
            """
        ),
        env=dict(
            BUILD_NAME=expected_build_name,
            BUILD_NUMBER=expected_build_number,
            BUILD_URL=expected_build_url,
        ),
    )

    assert result.exit_code == 0
    out, err = capfd.readouterr()
    sys.stdout.write(out)
    sys.stderr.write(err)

    build_name, build_number, build_url = out.splitlines()

    assert build_name == expected_build_name
    assert build_number == expected_build_number
    assert build_url == expected_build_url


def _timeout_mock_time_monotonic(clock_state):
    clock_state["time"] = clock_state.get("time", 0) + 1e-6
    return clock_state["time"]


def _timeout_mock_check_call(clock_state, args, *popenargs, timeout=None, **kwargs):
    cmd, delay = args
    assert cmd == "sleep"
    delay = float(delay)
    assert delay > 0
    if timeout is not None and delay > timeout:
        raise subprocess.TimeoutExpired(args, timeout)
    clock_state["time"] = clock_state.get("time", 0) + delay


@pytest.mark.parametrize("sleep", (0.002, 0.004, 0.006, 0.008), ids=lambda n: f"sleep={n}")
@pytest.mark.parametrize("timeout", (0.001, 0.003, 0.005, 0.007), ids=lambda n: f"timeout={n}")
def test_local_timeout(monkeypatch, run_hopic, sleep, timeout):
    clock_state = {}
    monkeypatch.setattr(time, "monotonic", functools.partial(_timeout_mock_time_monotonic, clock_state))
    monkeypatch.setattr(subprocess, "check_call", functools.partial(_timeout_mock_check_call, clock_state))

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""\
            phases:
              a:
                x:
                  - timeout: {timeout}
                    sh: sleep {sleep}
            """
        ),
    )
    if sleep < timeout:
        assert result.exception is None
    else:
        assert result.exception is not None
        if not isinstance(result.exception, StepTimeoutExpiredError):
            raise result.exception


def test_global_timeout_expire(monkeypatch, run_hopic):
    clock_state = {}
    monkeypatch.setattr(time, "monotonic", functools.partial(_timeout_mock_time_monotonic, clock_state))
    monkeypatch.setattr(subprocess, "check_call", functools.partial(_timeout_mock_check_call, clock_state))

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """\
            phases:
              a:
                x:
                  - timeout: 0.006
                  - timeout: 0.002
                    sh: sleep 0.001
                  - timeout: 0.002
                    sh: sleep 0.001
                  - sh: sleep 0.004
            """
        ),
    )
    assert isinstance(result.exception, StepTimeoutExpiredError)

    timeout_msg_re = re.compile(r"\brestrict.*?\bmax.*\bseconds?\b")
    timeout_msgs = tuple(msg for _, msg in result.logs if timeout_msg_re.search(msg))
    assert timeout_msgs, f"Didn't find any timeout related messages matching '{timeout_msg_re.pattern}'"

    assert "global" in timeout_msgs[-1], "timeout expiration wasn't caused by the _global_ timeout"


def test_variant_finally(monkeypatch, run_hopic):
    username = "test_username"
    password = "super_secret"
    credential_id = "test_credentialId"
    project_name = "test_project"

    expected_cmds = [("command", "a"), ("invalid", "cmd"), ("final", "one"), ("invalid", "cmd", "finally"), (username, password)]

    def get_credential_id(project_name_arg, cred_id):
        assert credential_id == cred_id
        assert project_name == project_name_arg
        return username, password

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected_cmds.pop(0)
        if "invalid" in args:
            raise subprocess.CalledProcessError(1, args)

    monkeypatch.setattr(credentials, "get_credential_by_id", get_credential_id)
    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            f"""
            project-name: {project_name}
            phases:
              a:
                x:
                  - sh: command a
                    finally:
                      - final one
                  - sh: invalid cmd
                    finally:
                      - invalid cmd finally
                  - sh: echo "never reach this point"
                    finally:
                      - "never execute finally"
                  - finally:
                    - with-credentials:
                        - id: {credential_id}
                          type: username-password
                          username-variable: 'TEST_USER'
                          password-variable: 'TEST_PASSWORD'
                    - $TEST_USER $TEST_PASSWORD
            """
        ),
    )
    assert result.exit_code != 0
    assert expected_cmds == []


def test_phase_variant_variables(monkeypatch, run_hopic):
    expected_cmds = [("command", "a", "x"), ("sh", "a", "x"), ("final", "a", "x")]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected_cmds.pop(0)

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """
            phases:
              a:
                x:
                  - sh: command $HOPIC_PHASE $HOPIC_VARIANT
                    finally:
                      - sh $HOPIC_PHASE $HOPIC_VARIANT
                  - finally:
                    - final $HOPIC_PHASE $HOPIC_VARIANT
            """
        ),
    )
    assert result.exit_code == 0
    assert expected_cmds == []


def test_chained_variants_finally(monkeypatch, run_hopic):
    expected_cmds = [("invalid", "command"), ("sh", "a", "x"), ("final", "a", "x"), ("chained", "b", "x")]

    def mock_check_call(args, *popenargs, **kwargs):
        assert tuple(args) == expected_cmds.pop(0)
        if "invalid" in args:
            raise subprocess.CalledProcessError(1, args)

    monkeypatch.setattr(subprocess, "check_call", mock_check_call)

    (result,) = run_hopic(
        ("build",),
        config=dedent(
            """
            phases:
              a:
                x:
                  - sh: invalid command
                    finally:
                      - sh $HOPIC_PHASE $HOPIC_VARIANT
                  - finally:
                    - final $HOPIC_PHASE $HOPIC_VARIANT

              b:
                x:
                  - wait-on-full-previous-phase: no
                  - sh: never execute!
                  - finally:
                    - chained $HOPIC_PHASE $HOPIC_VARIANT
            """
        ),
    )
    assert result.exit_code != 0
    assert expected_cmds == []



# Copyright (c) 2020 - 2020 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from .. import config_reader


def test_example(example_file):
    cfg = config_reader.read(
        example_file,
        {
            'WORKSPACE': '.',
            'CT_DEVENV_HOME': '/tools/devenv',
        },
    )

    json.dumps(cfg, cls=config_reader.JSONEncoder)



# Copyright (c) 2020 - 2020 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
import sys

try:
    import keyring
    import keyring.backends.kwallet
    import keyring.backends.OS_X
    import keyring.backends.SecretService
    import keyring.credentials
    import netstruct
    from contextlib import closing

    class OSXKeyring(keyring.backends.OS_X.Keyring):

        pattern = b'b$b$'

        def get_credential(self, service, username):
            if username is None:
                username = service
            cred = super().get_credential(service, username)
            if cred is not None:
                u, p = netstruct.unpack(self.pattern, cred.password.encode('UTF-8'))
                cred = keyring.credentials.SimpleCredential(u.decode('UTF-8'), p.decode('UTF-8'))
            return cred

        def set_password(self, service, username, password):
            password = netstruct.pack(self.pattern, username.encode('UTF-8'), password.encode('UTF-8')).decode('UTF-8')
            username = service
            return super().set_password(service, username, password)

    class KeePassKeyring(keyring.backends.SecretService.Keyring):
        """
        A version of the SecretService keyring API that falls back to using the camel case
        'UserName'. KeePass uses that as its attribute name instead of 'username' like most other
        implementations.
        """

        appid = 'Hopic'

        def get_credential(self, service, username):
            for attr in ('username', 'UserName'):
                query = {'service': service}
                if username:
                    query[attr] = username

                collection = self.get_preferred_collection()

                with closing(collection.connection):
                    items = collection.search_items(query)
                    for item in items:
                        self.unlock(item)
                        for attr in ('username', 'UserName'):
                            username = item.get_attributes().get(attr)
                            if username is not None:
                                break
                        return keyring.credentials.SimpleCredential(username, item.get_secret().decode('UTF-8'))

        def get_password(self, service, username):
            cred = self.get_credential(service, username)
            if cred is not None:
                return cred.password
except ImportError:
    keyring = None  # type: ignore # mypy is buggy for this try-except import style: https://github.com/python/mypy/issues/1153

_keyring_backend = None


def _init_keyring():
    if keyring is None:
        return

    global _keyring_backend
    if _keyring_backend is not None:
        return _keyring_backend

    backends = [keyring.get_keyring()]
    try:
        backends = backends[0].backends
    except AttributeError:
        pass

    for i, backend in enumerate(backends):
        # Substitute our own KeePass compatible keyring
        if isinstance(backend, keyring.backends.SecretService.Keyring):
            backends[i] = KeePassKeyring()
            backend = backends

        if isinstance(backend, (
                keyring.backends.kwallet.DBusKeyring,
                keyring.backends.SecretService.Keyring,
                )):
            backend.appid = 'Hopic'

        if isinstance(backend, keyring.backends.OS_X.Keyring):
            backends[i] = OSXKeyring()
            backend = backends

    if len(backends) == 1:
        _keyring_backend = backends[0]
    else:
        _keyring_backend = keyring.get_keyring()
    return _keyring_backend


def get_credential_by_id(project_name, cred_id):
    backend = _init_keyring()
    if backend is None:
        return None

    cred_name = f"{project_name}-{cred_id}"
    kcred = backend.get_credential(cred_name, None)
    if kcred is not None:
        return kcred.username, kcred.password
    elif hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
        username =           input(f"Username for {cred_name}: ")  # noqa: E222
        password = getpass.getpass(f"Password for {cred_name}: ")
        backend.set_password(cred_name, username, password)
        return username, password



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functionality for Hopic templates.
"""

from collections.abc import Sequence
import sys
from typing import (
    Any,
    Iterable,
    List,
    Tuple,
    Union,
)


def _kebabify(name: str) -> str:
    """Convert from snake_case to kebab-case"""
    return name.replace("_", "-")


def _name_to_arg(name: str) -> str:
    if name.startswith("_") and len(name) > 1:
        name = name[1:]
    if len(name) == 1:
        assert name not in ("_", "-")
        return f"-{name}"
    else:
        return f"--{_kebabify(name)}"


def _kwarg_to_arg(name: str, value: Any) -> Iterable[str]:
    if value is True:
        yield _name_to_arg(name)
    elif value is not False and value is not None:
        yield _name_to_arg(name)
        yield str(value)


def _kwargs_to_args(**kwargs: Any) -> Iterable[str]:
    for key, val in kwargs.items():
        if not isinstance(val, str) and isinstance(val, Sequence):
            for subval in val:
                yield from _kwarg_to_arg(key, subval)
        else:
            yield from _kwarg_to_arg(key, val)


def command(
    # NOTE: the_command_list param name is chosen to decrease probability of clashing with 'kwargs'
    the_command_list: Union[str, List[str], Tuple[str, ...]],
    *args: Any,
    **kwargs: Any,
):
    """
    Constructs an argument list from a Pythonic set of arguments and keyword arguments for executing
    some external program.
    """
    if isinstance(the_command_list, str):
        the_command_list = (the_command_list,)

    args = tuple(str(arg) for arg in args)
    if any(arg.startswith("-") for arg in args):
        # Prevent interpretation as an option
        args = ("--", *args)

    return (*the_command_list, *_kwargs_to_args(**kwargs), *args)


def module_command(
    # NOTE: the_module_and_command param name is chosen to decrease probability of clashing with 'kwargs'
    the_module_and_command: Union[str, List[str], Tuple[str, ...]],
    *args: Any,
    **kwargs: Any,
):
    """
    Constructs an argument list from a Pythonic set of arguments and keyword arguments for executing
    an executable Python module with.
    """
    if isinstance(the_module_and_command, str):
        the_module_and_command = (the_module_and_command,)

    return command((sys.executable, "-m", *the_module_and_command), *args, **kwargs)



# Copyright (c) 2021 - 2021 TomTom N.V. (https://tomtom.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import keyword
from pathlib import Path
import sys

import pytest

from hopic.template.utils import (
    command,
    module_command,
)


def test_args_only():
    assert command("cmd") == ("cmd",)
    assert command("cmd", 1) == ("cmd", "1")
    assert command("cmd", "a") == ("cmd", "a")
    assert command("cmd", 1, 2, 3, "a", "b", "c") == ("cmd", "1", "2", "3", "a", "b", "c")


def test_bool_kwargs_only():
    assert command("cmd", first=True, second=True) == ("cmd", "--first", "--second")
    assert command("cmd", second=True, first=True) == ("cmd", "--second", "--first")
    assert command("cmd", first=False, second=True) == ("cmd", "--second")
    assert command("cmd", first=True, second=False) == ("cmd", "--first")


def test_list_kwargs():
    assert command("cmd", puppet=("Woody", "Buzz")) == ("cmd", "--puppet", "Woody", "--puppet", "Buzz")
    assert command("cmd", puppet=["Buzz", "Woody"]) == ("cmd", "--puppet", "Buzz", "--puppet", "Woody")


def test_short_kwargs():
    assert command("cmd", x=True, v=True, f=Path("some/archive.tar")) == ("cmd", "-x", "-v", "-f", "some/archive.tar")
    assert command("cmd", f=(True, True), v=True) == ("cmd", "-f", "-f", "-v")


def test_str_coercion():
    assert command("cmd", Path("over/here.txt"), include=(Path("somewhere/else.txt"), Path("some/other/place.txt"))) == (
        "cmd",
        *("--include", "somewhere/else.txt"),
        *("--include", "some/other/place.txt"),
        "over/here.txt",
    )

    assert command("cmd", 0, arg=1, args=(2, True)) == ("cmd", "--arg", "1", "--args", "2", "--args", "0")


@pytest.mark.parametrize("keyword", keyword.kwlist)
def test_keyword_opts(keyword):
    if keyword.startswith("_"):
        pytest.skip("Ignoring keywords starting with underscores as we don't support those currently.")

    try:
        call = compile(f"command('cmd', {keyword}=True)", filename="", mode="eval")
    except SyntaxError:
        call = compile(f"command('cmd', _{keyword}=True)", filename="", mode="eval")

    assert eval(call) == ("cmd", f"--{keyword}")


def test_opt_like_args():
    assert command("cmd", Path("-cons.txt"), Path("+pros.txt"), evaluate=True) == ("cmd", "--evaluate", "--", "-cons.txt", "+pros.txt")


def test_list_command():
    assert command(("git", "add"), Path("a.txt"), Path("b.txt"), u=True) == ("git", "add", "-u", "a.txt", "b.txt")


def test_module_command():
    assert module_command(__name__, 1, 2, verbose=True) == (sys.executable, "-m", "hopic.test.test_template_utils", "--verbose", "1", "2")
    assert module_command((__name__, "sub"), 1, verbose=True) == (sys.executable, "-m", "hopic.test.test_template_utils", "sub", "--verbose", "1")



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import sgr_re
from .. import execution
import hopic
import click
import copy
import subprocess


def test_echo_cmd_dry_run_argument_parsing(monkeypatch):
    expected_cmd = ['echo', 'Bob the builder']
    expected_args = ['a', 'b', 'c']
    expected_kwargs = {'d': 'test string', 'e': 'different string'}

    def mock_check_call(arg, *args, **kwargs):
        assert arg == expected_cmd
        args_array = list(args)
        for element in args:
            assert expected_args.pop(0) in args
            args_array.remove(element)
        assert args_array == []

        for kwarg in copy.deepcopy(expected_kwargs):
            assert kwarg in kwargs
            assert kwargs[kwarg] == expected_kwargs[kwarg]
            del expected_kwargs[kwarg]

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    execution.echo_cmd(subprocess.check_call, expected_cmd, *expected_args, dry_run=False, **expected_kwargs)
    assert expected_args == []
    assert expected_kwargs == {}


def test_echo_cmd_dry_run(caplog, monkeypatch):
    expected_cmd = ['echo']
    expected_cmd_args = ['-e', 'Bob the builder']

    def mock_check_call(arg, *args, **kwargs):
        assert False

    monkeypatch.setattr(subprocess, 'check_call', mock_check_call)
    caplog.set_level("INFO")
    execution.echo_cmd(subprocess.check_call, expected_cmd + expected_cmd_args, **{'a': 'aa', 'b': 'bbb'}, dry_run=True)
    msgs = tuple(sgr_re.sub("", record.getMessage()) for record in caplog.records)
    assert "echo -e 'Bob the builder'" in msgs


def test_echo_cmd_return_value():
    def mock_executor(arg, *args, **kwargs):
        assert arg == 'command'
        return args[0]

    ctx = click.Context(hopic.cli.getinfo)
    with ctx:
        assert execution.echo_cmd_click(mock_executor, 'command', '42') == '42'
        assert execution.echo_cmd_click(mock_executor, 'command', b'42') == '42'



# Copyright (c) 2021 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

# Two things to note about this Python version-dependent import:
#  - importlib.metadata was introduced in 3.8
#  - importlib.metadata.entry_points() accepts the group keyword since 3.10

if sys.version_info[:2] >= (3, 10):
    from importlib import metadata as metadata  # noqa: F401
else:
    import importlib_metadata as metadata  # noqa: F401



# Copyright (c) 2020 - 2021 TomTom N.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from textwrap import dedent

import git
import pytest

from .. import config_reader

_git_time = f"{42 * 365 * 24 * 3600} +0000"
_author = git.Actor('Bob Tester', 'bob@example.net')
_commitargs = dict(
        author_date=_git_time,
        commit_date=_git_time,
        author=_author,
        committer=_author,
    )


@pytest.mark.parametrize('version_build', ('1.2.3', None))
@pytest.mark.parametrize('version_file', ('revision.txt', None))
def test_conventional_bump(version_build, version_file, run_hopic):
    config_reader.get_entry_points.cache_clear()  # clear (potential) dirty entry_point cache from previous tests as comissery is not installed in this test
    init_version = f'0.0.0+{version_build}' if version_build else '0.0.0'
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = 'hopic-ci-config.yaml'

        with (run_hopic.toprepo / cfg_file).open('w') as fp:
            fp.write(dedent(f"""\
                    version:
                      tag: yes
                      format: semver
                      bump:
                        policy: conventional-commits
                        strict: yes
                        on-every-change: no
                    {('  file: ' + version_file) if version_file else ''}
                    {('  build: ' + version_build) if version_build else ''}

                    phases:
                      style:
                        commit-messages: !template "commisery"
                    """))
        if version_file:
            with (run_hopic.toprepo / version_file).open('w') as fp:
                fp.write(f'version={init_version}\n')
            repo.index.add((version_file,))

        repo.index.add((cfg_file,))
        base_commit = repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag(init_version)

        base_commit = repo.index.commit(message='Invalid For Conventional Commits', **_commitargs)

        # PR branch
        repo.head.reference = repo.create_head('something-useful', base_commit)
        assert not repo.head.is_detached
        repo.head.reset(index=True, working_tree=True)

        repo.index.commit(message='fix: something useful', **_commitargs)

        repo.git.checkout('master')
        repo.git.merge('something-useful', no_ff=True, no_commit=True, env={
            'GIT_COMMITTER_NAME': 'My Name is Nobody',
            'GIT_COMMITTER_EMAIL': 'nobody@example.com',
        })
        repo.index.commit(message='Merge #1: feat: something useful', **_commitargs)

        expected_version = '0.1.0'
        expected_tag = expected_version + (f'+{version_build}' if version_build else '')

        # Make sure we're not on master: it would make the 'git push' from 'hopic submit' fail
        repo.git.checkout('something-useful')

    # Successful checkout and bump
    results = list(run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                '--author-name', _author.name,
                '--author-email', _author.email,
                'bump-version'),
            ('build',),
            ('submit',),
        ))

    assert all(result.exit_code == 0 for result in results)
    assert results[1].stdout.splitlines()[-1].split('+')[0] == expected_version

    with git.Repo(run_hopic.toprepo) as repo:
        # Switch back to master to be able to easily look at its contents
        repo.git.checkout('master')

        assert repo.tags[expected_tag].commit == repo.head.commit
        if version_file:
            with (run_hopic.toprepo / version_file).open('r') as fp:
                assert fp.read() == f'version={expected_version}\n'


def test_bump_skipped_when_no_new_commits(run_hopic):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = 'hopic-ci-config.yaml'

        with (run_hopic.toprepo / cfg_file).open('w') as fp:
            fp.write(dedent("""\
                    version:
                      tag: yes
                      format: semver
                      bump:
                        policy: conventional-commits
                        strict: yes
                        on-every-change: no

                    phases:
                      style:
                        commit-messages: !template "commisery"
                    """))
        repo.index.add((cfg_file,))
        repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag('0.0.0')

    # Successful checkout and bump
    *_, result = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                '--author-name', _author.name,
                '--author-email', _author.email,
                'bump-version'),
        )

    assert result.exit_code == 0
    assert result.stdout == ''
    assert (logging.INFO, "Not bumping because no new commits are present since the last tag '0.0.0'") in result.logs


def test_bump_skipped_when_no_bumpable_commits(run_hopic):
    with git.Repo.init(run_hopic.toprepo, expand_vars=False) as repo:
        cfg_file = 'hopic-ci-config.yaml'

        with (run_hopic.toprepo / cfg_file).open('w') as fp:
            fp.write(dedent("""\
                    version:
                      tag: yes
                      format: semver
                      bump:
                        policy: conventional-commits
                        strict: yes
                        on-every-change: no

                    phases:
                      style:
                        commit-messages: !template "commisery"
                    """))
        repo.index.add((cfg_file,))
        repo.index.commit(message='Initial commit', **_commitargs)
        repo.git.branch('master', move=True)
        repo.create_tag('0.0.0')

        repo.index.commit(message='ci: bla bla', **_commitargs)

    # Successful checkout and bump
    *_, result = run_hopic(
            ('checkout-source-tree', '--target-remote', run_hopic.toprepo, '--target-ref', 'master'),
            ('prepare-source-tree',
                '--author-date', f"@{_git_time}",
                '--commit-date', f"@{_git_time}",
                '--author-name', _author.name,
                '--author-email', _author.email,
                'bump-version'),
        )

    assert result.exit_code == 0
    assert result.stdout == ''
    assert (
        logging.ERROR,
        "Version bumping requested, but the version policy 'conventional-commits' decided not to bump from '0.0.1-1+gee3642c057a2af'",
    ) in result.logs
